# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import copy
import hashlib
import json

import eventlet
import mock

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils
from congress import exception
from congress.tests import base
from congress.tests.datasources import util
from congress.tests import helper


class TestDatasourceDriver(base.TestCase):

    def setUp(self):
        super(TestDatasourceDriver, self).setUp()
        self.val_trans = {'translation-type': 'VALUE'}

    def compute_hash(self, obj):
        s = json.dumps(sorted(obj, key=(lambda x: str(type(x)) + repr(x))),
                       sort_keys=True)
        h = hashlib.md5(s.encode('ascii')).hexdigest()
        return h

    def test_translator_key_elements(self):
        """Test for keys of all translator."""
        expected_params = {
            'hdict': ('translation-type', 'table-name', 'parent-key',
                      'id-col', 'selector-type', 'field-translators',
                      'in-list', 'parent-col-name', 'objects-extract-fn',
                      'parent-key-desc'),
            'vdict': ('translation-type', 'table-name', 'parent-key',
                      'id-col', 'key-col', 'val-col', 'translator',
                      'parent-col-name', 'objects-extract-fn'),
            'list': ('translation-type', 'table-name', 'parent-key',
                     'id-col', 'val-col', 'translator', 'parent-col-name',
                     'objects-extract-fn', 'parent-key-desc', 'val-col-desc'),
            }

        actual_params = {
            'hdict': datasource_driver.DataSourceDriver.HDICT_PARAMS,
            'vdict': datasource_driver.DataSourceDriver.VDICT_PARAMS,
            'list': datasource_driver.DataSourceDriver.LIST_PARAMS,
            }

        for key, params in actual_params.items():
            expected = expected_params[key]
            self.assertEqual(params, expected)

    def test_in_list_results_hdict_hdict(self):
        ports_fixed_ips_translator = {
            'translation-type': 'HDICT',
            'table-name': 'fixed-ips',
            'parent-key': 'id',
            'selector-type': 'DICT_SELECTOR',
            'in-list': True,
            'field-translators':
                ({'fieldname': 'ip_address', 'translator': self.val_trans},
                 {'fieldname': 'subnet_id', 'translator': self.val_trans})}

        ports_translator = {
            'translation-type': 'HDICT',
            'table-name': 'ports',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'id', 'translator': self.val_trans},
                 {'fieldname': 'fixed_ips',
                  'translator': ports_fixed_ips_translator})}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        driver.register_translator(ports_translator)
        ports = [{'id': '12345',
                  'fixed_ips': [{'ip_address': '1.1.1.1', 'subnet_id': 'aa'},
                                {'ip_address': '2.2.2.2', 'subnet_id': 'bb'}]}]
        row_data = driver.convert_objs(ports, ports_translator)
        expected = [('fixed-ips', ('12345', '1.1.1.1', 'aa')),
                    ('fixed-ips', ('12345', '2.2.2.2', 'bb')),
                    ('ports', ('12345',))]
        self.assertEqual(row_data, expected)

    def test_getting_parent_key_from_nested_tables(self):
        level3_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level3',
            'parent-key': 'parent_key',
            'selector-type': 'DICT_SELECTOR',
            'in-list': True,
            'field-translators':
                ({'fieldname': 'level3_thing', 'translator': self.val_trans},)}

        level2_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level2',
            'parent-key': 'id',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'thing', 'translator': self.val_trans},
                 {'fieldname': 'level3',
                  'translator': level3_translator})}

        level1_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level1',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'id', 'translator': self.val_trans},
                 {'fieldname': 'level2',
                  'translator': level2_translator})}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        driver.register_translator(level1_translator)
        data = [
            {'id': 11, 'level2':
                {'thing': 'blah!', 'level3': [{'level3_thing': '12345'}]}}]

        row_data = driver.convert_objs(data, level1_translator)
        expected = [('level3', (11, '12345')),
                    ('level2', (11, 'blah!')),
                    ('level1', (11,))]
        self.assertEqual(row_data, expected)

    def test_parent_col_name_in_hdict(self):
        level2_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level2',
            'parent-key': 'id',
            'parent-col-name': 'level1_id',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'thing', 'translator': self.val_trans},)}

        level1_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level1',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'id', 'translator': self.val_trans},
                 {'fieldname': 'level2',
                  'translator': level2_translator})}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        datasource_driver.DataSourceDriver.TRANSLATORS = [level1_translator]
        driver.register_translator(level1_translator)
        # test schema
        schema = driver.get_schema()
        expected = {'level1': ({'name': 'id', 'desc': None},),
                    'level2': ({'name': 'level1_id', 'desc': None},
                               {'name': 'thing', 'desc': None})}
        self.assertEqual(schema, expected)

        # test data
        data = [{'id': 11, 'level2': {'thing': 'blah!'}}]
        row_data = driver.convert_objs(data, level1_translator)
        expected = [('level2', (11, 'blah!')), ('level1', (11,))]
        self.assertEqual(row_data, expected)

    def test_parent_col_name_in_vdict(self):
        level2_translator = {
            'translation-type': 'VDICT',
            'table-name': 'level2',
            'parent-key': 'id',
            'key-col': 'id',
            'val-col': 'value',
            'parent-col-name': 'level1_id',
            'translator': self.val_trans}

        level1_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level1',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'id', 'translator': self.val_trans},
                 {'fieldname': 'level2',
                  'translator': level2_translator})}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        datasource_driver.DataSourceDriver.TRANSLATORS = [level1_translator]
        # test schema
        schema = driver.get_schema()
        expected = {'level1': ({'name': 'id', 'desc': None},),
                    'level2': ('level1_id', 'id', 'value')}
        self.assertEqual(expected, schema)

        # test data
        data = [{'id': 11, 'level2': {'thing': 'blah!'}}]
        row_data = driver.convert_objs(data, level1_translator)
        expected = [('level2', (11, 'thing', 'blah!')), ('level1', (11,))]
        self.assertEqual(row_data, expected)

    def test_parent_col_name_in_list(self):
        level2_translator = {
            'translation-type': 'LIST',
            'table-name': 'level2',
            'parent-key': 'id',
            'parent-key-desc': 'level1_parent-desc',
            'parent-col-name': 'level1_id',
            'val-col': 'level_1_data',
            'val-col-desc': 'level_1_desc',
            'translator': self.val_trans}

        level1_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level1',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'id', 'translator': self.val_trans},
                 {'fieldname': 'level2',
                  'translator': level2_translator})}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        datasource_driver.DataSourceDriver.TRANSLATORS = [level1_translator]
        # test schema
        schema = driver.get_schema()
        expected = {'level1': ({'name': 'id', 'desc': None},),
                    'level2': ({'name': 'level1_id',
                                'desc': 'level1_parent-desc'},
                               {'name': 'level_1_data',
                                'desc': 'level_1_desc'})}
        self.assertEqual(expected, schema)

        # test data
        data = [{'id': 11, 'level2': ['thing']}]
        row_data = driver.convert_objs(data, level1_translator)
        expected = [('level2', (11, 'thing')), ('level1', (11,))]
        self.assertEqual(row_data, expected)

    def test_check_for_duplicate_table_names_hdict_list(self):
        translator = {
            'translation-type': 'HDICT',
            'table-name': 'table1',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'tags',
                  'translator': {'translation-type': 'LIST',
                                 'table-name': 'table1',
                                 'val-col': 'tag',
                                 'translator': self.val_trans}},)}
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.DuplicateTableName,
                          driver.register_translator,
                          translator)

    def test_check_for_duplicate_table_names_nested_list_list(self):
        # Test a LIST containing a LIST with the same table name.
        translator = {'translation-type': 'LIST', 'table-name': 'testtable',
                      'id-col': 'id_col', 'val-col': 'value',
                      'translator': {'translation-type': 'LIST',
                                     'table-name': 'testtable',
                                     'id-col': 'id', 'val-col': 'val',
                                     'translator': self.val_trans}}
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.DuplicateTableName,
                          driver.register_translator, translator)

    def test_check_for_duplicate_table_names_in_different_translator(self):
        translator = {
            'translation-type': 'HDICT',
            'table-name': 'table1',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'tags',
                  'translator': {'translation-type': 'LIST',
                                 'table-name': 'table2',
                                 'val-col': 'tag',
                                 'translator': self.val_trans}},)}
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        driver.register_translator(translator)
        self.assertRaises(exception.DuplicateTableName,
                          driver.register_translator,
                          translator)

    def test_check_for_duplicate_table_names_hdict_hdict(self):
        translator = {
            'translation-type': 'HDICT',
            'table-name': 'table1',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'tags',
                  'translator': {'translation-type': 'HDICT',
                                 'table-name': 'table1',
                                 'selector-type': 'DICT_SELECTOR',
                                 'field-translators':
                                     ({'fieldname': 'x',
                                       'translator': self.val_trans},)}},)}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.DuplicateTableName,
                          driver.register_translator,
                          translator)

    def test_invalid_translation_type(self):
        translator = {'translation-type': 'YOYO',
                      'table-name': 'table1'}
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidTranslationType,
                          driver.register_translator,
                          translator)

        translator = {'translation-type': 'LIS',
                      'table-name': 'table1'}
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidTranslationType,
                          driver.register_translator,
                          translator)

    def test_no_key_col_in_vdict(self):
        translator = {'translation-type': 'VDICT',
                      'table-name': 'table1',
                      'val-col': 'id-col'}
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidParamException,
                          driver.register_translator,
                          translator)

    def test_no_val_col_in_vdict(self):
        translator = {'translation-type': 'VDICT',
                      'table-name': 'table1',
                      'key-col': 'id-col'}
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidParamException,
                          driver.register_translator,
                          translator)

    def test_no_val_col_in_list(self):
        translator = {'translation-type': 'LIST',
                      'table-name': 'table1'}
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidParamException,
                          driver.register_translator,
                          translator)

    def test_no_parent_key_id_col(self):
        translator = {'translation-type': 'LIST',
                      'table-name': 'table1',
                      'id-col': 'id-col',
                      'parent-key': 'parent_key_column'}

        # Test LIST
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidParamException,
                          driver.register_translator,
                          translator)
        # Test HDICT
        translator['translation-type'] = 'VDICT'
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidParamException,
                          driver.register_translator,
                          translator)
        # Test HDICT
        translator['translation-type'] = 'HDICT'
        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidParamException,
                          driver.register_translator,
                          translator)

    def test_check_no_extra_params(self):
        translator = {'translation-type': 'LIST',
                      'table-name': 'table1',
                      'id-col': 'id-col',
                      'invalid_column': 'blah'}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidParamException,
                          driver.register_translator,
                          translator)

    def test_check_no_extra_params_nested_hdict(self):
        translator = {
            'translation-type': 'HDICT',
            'table-name': 'table1',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'tags',
                  'translator': {'translation-type': 'HDICT',
                                 'table-name': 'table2',
                                 'selector-type': 'DICT_SELECTOR',
                                 'invalid_column': 'yaya',
                                 'field-translators':
                                     ({'fieldname': 'x',
                                       'translator': self.val_trans},)}},)}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidParamException,
                          driver.register_translator,
                          translator)

    def test_check_no_extra_params_nested_list_hdict(self):
        translator = {
            'translation-type': 'LIST',
            'table-name': 'table1',
            'val-col': 'fixed_ips',
            'translator': {
                'table-name': 'table2',
                'invalid-column': 'hello_there!',
                'translation-type': 'HDICT',
                'selector-type': 'DICT_SELECTOR',
                'field-translators':
                    ({'fieldname': 'ip_address',
                      'translator': self.val_trans},)}}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        self.assertRaises(exception.InvalidParamException,
                          driver.register_translator,
                          translator)

    def test_convert_vdict_with_id(self):
        # Test a single VDICT with an id column.
        resp = {'a': 'FOO', 'b': 123}
        translator = {'translation-type': 'VDICT', 'table-name': 'testtable',
                      'id-col': 'id_col', 'key-col': 'key', 'val-col': 'value',
                      'translator': self.val_trans}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        k1 = self.compute_hash((('a', 'FOO'), ('b', 123)))

        self.assertEqual(2, len(rows))
        self.assertEqual(k1, k)
        self.assertTrue(('testtable', (k, 'a', 'FOO')) in rows)
        self.assertTrue(('testtable', (k, 'b', 123)) in rows)

    def test_convert_vdict_without_id(self):
        # Test a single VDICT without an id column.
        resp = {'a': 'FOO', 'b': 123}
        translator = {'translation-type': 'VDICT', 'table-name': 'testtable',
                      'key-col': 'key', 'val-col': 'value',
                      'translator': self.val_trans}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(2, len(rows))
        self.assertIsNone(k)
        self.assertTrue(('testtable', ('a', 'FOO')) in rows)
        self.assertTrue(('testtable', ('b', 123)) in rows)

    def test_convert_vdict_with_id_function(self):
        # Test a single VDICT with an id column that is a function.
        resp = {'a': 'FOO', 'b': 123}
        translator = {'translation-type': 'VDICT', 'table-name': 'testtable',
                      'id-col': lambda obj: 'id:' + obj['a'],
                      'key-col': 'key', 'val-col': 'value',
                      'translator': self.val_trans}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        k1 = 'id:FOO'

        self.assertEqual(2, len(rows))
        self.assertEqual(k1, k)
        self.assertTrue(('testtable', (k, 'a', 'FOO')) in rows)
        self.assertTrue(('testtable', (k, 'b', 123)) in rows)

    def test_convert_vdict_list(self):
        # Test a VDICT that contains lists.
        resp = {'foo': (1, 2, 3), 'bar': ('a', 'b')}
        translator = {'translation-type': 'VDICT', 'table-name': 'testtable',
                      'id-col': 'id_col', 'key-col': 'key', 'val-col': 'value',
                      'translator': {'translation-type': 'LIST',
                                     'table-name': 'subtable',
                                     'id-col': 'id_col', 'val-col': 'val_col',
                                     'translator': self.val_trans}}
        rows, actual_k = datasource_driver.DataSourceDriver.convert_obj(
            resp, translator)

        k1 = self.compute_hash((1, 2, 3))
        k2 = self.compute_hash(('a', 'b'))
        k = self.compute_hash((('foo', k1), ('bar', k2)))

        self.assertEqual(7, len(rows))
        self.assertEqual(k, actual_k)

        self.assertTrue(('subtable', (k1, 1)) in rows)
        self.assertTrue(('subtable', (k1, 2)) in rows)
        self.assertTrue(('subtable', (k1, 3)) in rows)
        self.assertTrue(('subtable', (k2, 'a')) in rows)
        self.assertTrue(('subtable', (k2, 'b')) in rows)
        self.assertTrue(('testtable', (k, 'foo', k1)) in rows)
        self.assertTrue(('testtable', (k, 'bar', k2)) in rows)

    def test_convert_vdict_is_none(self):
        # Test a single VDICT with an id column.
        resp = None
        translator = {'translation-type': 'VDICT', 'table-name': 'testtable',
                      'id-col': 'id_col', 'key-col': 'key', 'val-col': 'value',
                      'translator': self.val_trans}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)
        self.assertTrue(rows is None)

    def test_convert_list_with_id(self):
        # Test a single LIST with an id_column
        resp = (1, 'a', 'b', True)
        translator = {'translation-type': 'LIST', 'table-name': 'testtable',
                      'id-col': 'id_col', 'val-col': 'value',
                      'translator': self.val_trans}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        k1 = self.compute_hash((1, 'a', 'b', 'True'))

        self.assertEqual(4, len(rows))
        self.assertEqual(k1, k)
        self.assertTrue(('testtable', (k, 1)) in rows)
        self.assertTrue(('testtable', (k, 'a')) in rows)
        self.assertTrue(('testtable', (k, 'b')) in rows)
        self.assertTrue(('testtable', (k, 'True')) in rows)

    def test_convert_list_with_id_function(self):
        # Test a single LIST with an id function
        resp = (1, 'a', 'b', True)
        translator = {'translation-type': 'LIST', 'table-name': 'testtable',
                      'id-col': lambda obj: obj[0], 'val-col': 'value',
                      'translator': self.val_trans}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        k1 = 1

        self.assertEqual(4, len(rows))
        self.assertEqual(k1, k)
        self.assertTrue(('testtable', (k, 1)) in rows)
        self.assertTrue(('testtable', (k, 'a')) in rows)
        self.assertTrue(('testtable', (k, 'b')) in rows)
        self.assertTrue(('testtable', (k, 'True')) in rows)

    def test_convert_list_without_id(self):
        # Test a single LIST without an id_column
        resp = (1, 'a', 'b', True)
        translator = {'translation-type': 'LIST', 'table-name': 'testtable',
                      'val-col': 'value', 'translator': self.val_trans}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(4, len(rows))
        self.assertIsNone(k)
        self.assertTrue(('testtable', (1,)) in rows)
        self.assertTrue(('testtable', ('a',)) in rows)
        self.assertTrue(('testtable', ('b',)) in rows)
        self.assertTrue(('testtable', ('True',)) in rows)

    def test_convert_list_with_sublist(self):
        # Test a single LIST with an id_column
        resp = ((1, 2, 3), ('a', 'b', 'c'), (True, False))
        translator = {'translation-type': 'LIST', 'table-name': 'testtable',
                      'id-col': 'id_col', 'val-col': 'value',
                      'translator': {'translation-type': 'LIST',
                                     'table-name': 'subtable',
                                     'id-col': 'id_col', 'val-col': 'val_col',
                                     'translator': self.val_trans}}
        rows, actual_k = datasource_driver.DataSourceDriver.convert_obj(
            resp, translator)

        k1 = self.compute_hash((1, 2, 3))
        k2 = self.compute_hash(('a', 'b', 'c'))
        k3 = self.compute_hash(('True', 'False'))
        k = self.compute_hash((k1, k2, k3))

        self.assertEqual(11, len(rows))
        self.assertEqual(k, actual_k)
        self.assertTrue(('subtable', (k1, 1)) in rows)
        self.assertTrue(('subtable', (k1, 2)) in rows)
        self.assertTrue(('subtable', (k1, 3)) in rows)
        self.assertTrue(('subtable', (k2, 'a')) in rows)
        self.assertTrue(('subtable', (k2, 'b')) in rows)
        self.assertTrue(('subtable', (k2, 'c')) in rows)
        self.assertTrue(('subtable', (k3, 'True')) in rows)
        self.assertTrue(('subtable', (k3, 'False')) in rows)
        self.assertTrue(('testtable', (k, k1)) in rows)
        self.assertTrue(('testtable', (k, k2)) in rows)
        self.assertTrue(('testtable', (k, k3)) in rows)

    def test_convert_recursive_hdict_single_fields(self):
        # Test simple fields inside of an HDICT
        resp = util.ResponseObj({'testfield1': 'FOO',
                                 'testfield2': 123})
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DOT_SELECTOR',
                      'field-translators': (
                          {'fieldname': 'testfield1', 'col': 'col1',
                           'translator': self.val_trans},
                          {'fieldname': 'testfield2', 'col': 'col2',
                           'translator': self.val_trans})}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(1, len(rows))
        self.assertIsNone(k)
        self.assertEqual([('testtable', ('FOO', 123))], rows)

    def test_convert_recursive_hdict_single_fields_empty_fields(self):
        # Test simple fields inside of an HDICT where the translator
        # interprests a non-present field as None.
        resp = util.ResponseObj({'testfield1': 'FOO'})
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DOT_SELECTOR',
                      'field-translators': (
                          {'fieldname': 'testfield1', 'col': 'col1',
                           'translator': self.val_trans},
                          {'fieldname': 'testfield2', 'col': 'col2',
                           'translator': self.val_trans})}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(1, len(rows))
        self.assertIsNone(k)
        self.assertEqual([('testtable', ('FOO', 'None'))], rows)

    def test_convert_recursive_hdict_single_fields_default_col(self):
        # Test simple fields inside of an HDICT using the default col name.

        resp = util.ResponseObj({'testfield1': 'FOO'})
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DOT_SELECTOR',
                      'field-translators': (
                          {'fieldname': 'testfield1',
                           'translator': self.val_trans},)}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(1, len(rows))
        self.assertIsNone(k)
        self.assertEqual([('testtable', ('FOO',))], rows)

    def test_convert_recursive_hdict_extract_subfields(self):
        # Test simple fields inside of an HDICT
        # Also tests with and without extract-fn.
        field = util.ResponseObj({'b': 123})
        resp = util.ResponseObj({'testfield1': {'a': 'FOO'},
                                 'testfield2': field,
                                 'testfield3': 456})
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'id-col': 'id_col', 'selector-type': 'DOT_SELECTOR',
                      'field-translators': (
                          {'fieldname': 'testfield1', 'col': 'col1',
                           'translator': {'translation-type': 'VALUE',
                                          'extract-fn': lambda x: x['a']}},
                          {'fieldname': 'testfield2', 'col': 'col2',
                           'translator': {'translation-type': 'VALUE',
                                          'extract-fn': lambda x: x.b}},
                          {'fieldname': 'testfield3', 'col': 'col3',
                           'translator': {'translation-type': 'VALUE'}})}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(1, len(rows))
        self.assertEqual(self.compute_hash(('FOO', 123, 456)), k)
        self.assertEqual([('testtable', (k, 'FOO', 123, 456))], rows)

    def test_convert_recursive_hdict_sublists(self):
        # Test sublists inside of an HDICT
        resp = util.ResponseObj({'testfield1': ('FOO', 'BAR'),
                                 'testfield2': (1, 2, 3)})
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DOT_SELECTOR',
                      'field-translators': (
                          {'fieldname': 'testfield1', 'col': 'col1',
                           'translator': {'translation-type': 'LIST',
                                          'table-name': 'subtable1',
                                          'id-col': 'id', 'val-col': 'value',
                                          'translator': self.val_trans}},
                          {'fieldname': 'testfield2', 'col': 'col2',
                           'translator': {'translation-type': 'LIST',
                                          'table-name': 'subtable2',
                                          'id-col': 'id', 'val-col': 'value',
                                          'translator': self.val_trans}})}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        k1 = self.compute_hash(('FOO', 'BAR'))
        k2 = self.compute_hash((1, 2, 3))

        self.assertIsNone(k)
        self.assertEqual(6, len(rows))
        self.assertTrue(('subtable1', (k1, 'FOO')) in rows)
        self.assertTrue(('subtable1', (k1, 'BAR')) in rows)
        self.assertTrue(('subtable2', (k2, 1)) in rows)
        self.assertTrue(('subtable2', (k2, 2)) in rows)
        self.assertTrue(('subtable2', (k2, 3)) in rows)
        self.assertTrue(('testtable', (k1, k2)) in rows)

    def test_convert_recursive_hdict_vdict(self):
        # Test translator of an VDICT inside of an HDICT
        resp = util.ResponseObj({'testfield1': {'a': 123, 'b': 456},
                                 'testfield2': {'c': 'abc', 'd': 'def'}})
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DOT_SELECTOR',
                      'field-translators': (
                          {'fieldname': 'testfield1', 'col': 'col1',
                           'translator': {'translation-type': 'VDICT',
                                          'table-name': 'subtable1',
                                          'id-col': 'id', 'key-col': 'key',
                                          'val-col': 'value',
                                          'translator': self.val_trans}},
                          {'fieldname': 'testfield2', 'col': 'col2',
                           'translator': {'translation-type': 'VDICT',
                                          'table-name': 'subtable2',
                                          'id-col': 'id', 'key-col': 'key',
                                          'val-col': 'value',
                                          'translator': self.val_trans}})}
        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        k1 = self.compute_hash((('a', 123), ('b', 456)))
        k2 = self.compute_hash((('c', 'abc'), ('d', 'def')))

        self.assertIsNone(k)
        self.assertEqual(5, len(rows))
        self.assertTrue(('subtable1', (k1, 'a', 123)) in rows)
        self.assertTrue(('subtable1', (k1, 'b', 456)) in rows)
        self.assertTrue(('subtable2', (k2, 'c', 'abc')) in rows)
        self.assertTrue(('subtable2', (k2, 'd', 'def')) in rows)
        self.assertTrue(('testtable', (k1, k2)) in rows)

    def test_convert_recursive_hdict_hdict(self):
        # Test translator of an HDICT inside of an HDICT.
        resp = util.ResponseObj({'testfield1': {'a': 123, 'b': 456},
                                 'testfield2': {'c': 'abc', 'd': 'def'}})

        subtranslator_1 = {'translation-type': 'HDICT',
                           'table-name': 'subtable1',
                           'selector-type': 'DICT_SELECTOR',
                           'id-col': 'id',
                           'field-translators': (
                               {'fieldname': 'a',
                                'col': 'a1',
                                'translator': self.val_trans},
                               {'fieldname': 'b',
                                'col': 'b1',
                                'translator': self.val_trans})}

        subtranslator_2 = {'translation-type': 'HDICT',
                           'table-name': 'subtable2',
                           'selector-type': 'DICT_SELECTOR',
                           'id-col': 'id',
                           'field-translators': (
                               {'fieldname': 'c',
                                'col': 'c1',
                                'translator': self.val_trans},
                               {'fieldname': 'd',
                                'col': 'd1',
                                'translator': self.val_trans})}

        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DOT_SELECTOR',
                      'field-translators': (
                          {'fieldname': 'testfield1', 'col': 'col1',
                           'translator': subtranslator_1},
                          {'fieldname': 'testfield2', 'col': 'col2',
                           'translator': subtranslator_2})}

        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        k1 = self.compute_hash((123, 456))
        k2 = self.compute_hash(('abc', 'def'))

        self.assertIsNone(k)
        self.assertEqual(3, len(rows))
        self.assertTrue(('subtable1', (k1, 123, 456)) in rows)
        self.assertTrue(('subtable2', (k2, 'abc', 'def')) in rows)
        self.assertTrue(('testtable', (k1, k2)) in rows)

    def test_convert_hdict_hdict_parent_key_without_id(self):
        # Test a HDICT that contains lists using a parent_key.
        resp = {'foreign-key': 100, 'foo': {'f1': 123}}
        subtranslator = {'translation-type': 'HDICT',
                         'table-name': 'subtable',
                         'parent-key': 'foreign-key',
                         'selector-type': 'DICT_SELECTOR',
                         'field-translators': ({'fieldname': 'f1',
                                                'translator': self.val_trans},
                                               )}
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DICT_SELECTOR',
                      'field-translators': ({'fieldname': 'foreign-key',
                                             'translator': self.val_trans},
                                            {'fieldname': 'foo',
                                             'translator': subtranslator})}

        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(2, len(rows))
        self.assertIsNone(k)

        self.assertTrue(('subtable', (100, 123)) in rows)
        self.assertTrue(('testtable', (100,)) in rows)

    def test_convert_hdict_hdict_parent_key_with_id(self):
        # Test a HDICT that contains lists using a parent_key.
        resp = {'foreign-key': 100, 'foo': {'f1': 123}}
        subtranslator = {'translation-type': 'HDICT',
                         'table-name': 'subtable',
                         'parent-key': 'foreign-key',
                         'selector-type': 'DICT_SELECTOR',
                         'field-translators': ({'fieldname': 'f1',
                                                'translator': self.val_trans},
                                               )}
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DICT_SELECTOR', 'id-col': 'id',
                      'field-translators': ({'fieldname': 'foreign-key',
                                             'translator': self.val_trans},
                                            {'fieldname': 'foo',
                                             'translator': subtranslator})}

        rows, actual_k = datasource_driver.DataSourceDriver.convert_obj(
            resp, translator)

        k = self.compute_hash((100,))
        self.assertEqual(2, len(rows))
        self.assertEqual(k, actual_k)

        self.assertTrue(('subtable', (100, 123)) in rows)
        self.assertTrue(('testtable', (k, 100,)) in rows)

    def test_convert_hdict_vdict_parent_key_without_id(self):
        # Test a HDICT that contains lists using a parent_key.
        resp = util.ResponseObj({'foreign-key': 100,
                                 'foo': {'f1': 123, 'f2': 456}})
        subtranslator = {'translation-type': 'VDICT',
                         'table-name': 'subtable',
                         'parent-key': 'foreign-key',
                         'key-col': 'key_col',
                         'val-col': 'val_col',
                         'translator': self.val_trans}
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DOT_SELECTOR',
                      'field-translators': ({'fieldname': 'foreign-key',
                                             'translator': self.val_trans},
                                            {'fieldname': 'foo',
                                             'translator': subtranslator})}

        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(3, len(rows))
        self.assertIsNone(k)

        self.assertTrue(('subtable', (100, 'f1', 123)) in rows)
        self.assertTrue(('subtable', (100, 'f2', 456)) in rows)
        self.assertTrue(('testtable', (100,)) in rows)

    def test_convert_hdict_vdict_parent_key_with_id(self):
        # Test a HDICT that contains lists using a parent_key.
        resp = util.ResponseObj({'foreign-key': 100,
                                 'foo': {'f1': 123, 'f2': 456}})
        list_translator = {'translation-type': 'VDICT',
                           'table-name': 'subtable',
                           'parent-key': 'foreign-key',
                           'key-col': 'key_col',
                           'val-col': 'val_col',
                           'translator': self.val_trans}
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'id-col': 'id', 'selector-type': 'DOT_SELECTOR',
                      'field-translators': ({'fieldname': 'foreign-key',
                                             'translator': self.val_trans},
                                            {'fieldname': 'foo',
                                             'translator': list_translator})}

        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(3, len(rows))
        self.assertEqual(self.compute_hash((100,)), k)

        self.assertTrue(('subtable', (100, 'f1', 123)) in rows)
        self.assertTrue(('subtable', (100, 'f2', 456)) in rows)
        self.assertTrue(('testtable', (k, 100)) in rows)

    def test_convert_hdict_list_parent_key_without_id(self):
        # Test a HDICT that contains lists using a parent_key.
        resp = util.ResponseObj({'foreign-key': 100, 'foo': (1, 2)})
        list_translator = {'translation-type': 'LIST',
                           'table-name': 'subtable',
                           'parent-key': 'foreign-key',
                           'val-col': 'val_col',
                           'translator': self.val_trans}
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'selector-type': 'DOT_SELECTOR',
                      'field-translators': ({'fieldname': 'foreign-key',
                                             'translator': self.val_trans},
                                            {'fieldname': 'foo',
                                             'translator': list_translator})}

        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(3, len(rows))
        self.assertIsNone(k)

        self.assertTrue(('subtable', (100, 1)) in rows)
        self.assertTrue(('subtable', (100, 2)) in rows)
        self.assertTrue(('testtable', (100,)) in rows)

    def test_convert_hdict_list_parent_key_with_id(self):
        # Test a HDICT that contains lists using a parent_key.
        resp = util.ResponseObj({'foreign-key': 100, 'foo': (1, 2)})
        list_translator = {'translation-type': 'LIST',
                           'table-name': 'subtable',
                           'parent-key': 'foreign-key',
                           'val-col': 'val_col',
                           'translator': self.val_trans}
        translator = {'translation-type': 'HDICT', 'table-name': 'testtable',
                      'id-col': 'id', 'selector-type': 'DOT_SELECTOR',
                      'field-translators': ({'fieldname': 'foreign-key',
                                             'translator': self.val_trans},
                                            {'fieldname': 'foo',
                                             'translator': list_translator})}

        rows, k = datasource_driver.DataSourceDriver.convert_obj(resp,
                                                                 translator)

        self.assertEqual(3, len(rows))
        self.assertEqual(self.compute_hash((100,)), k)

        self.assertTrue(('subtable', (100, 1)) in rows)
        self.assertTrue(('subtable', (100, 2)) in rows)
        self.assertTrue(('testtable', (k, 100)) in rows)

    def test_convert_vdict_list_parent_key_without_id(self):
        # Test a VDICT that contains lists using a parent_key.
        resp = {'foo': (1, 2, 3), 'bar': ('a', 'b')}
        translator = {'translation-type': 'VDICT', 'table-name': 'testtable',
                      'key-col': 'key', 'val-col': 'value',
                      'translator': {'translation-type': 'LIST',
                                     'table-name': 'subtable',
                                     'parent-key': 'key',
                                     'val-col': 'val_col',
                                     'translator': self.val_trans}}
        rows, actual_k = datasource_driver.DataSourceDriver.convert_obj(
            resp, translator)

        self.assertEqual(7, len(rows))
        self.assertIsNone(actual_k)

        self.assertTrue(('subtable', ('foo', 1)) in rows)
        self.assertTrue(('subtable', ('foo', 2)) in rows)
        self.assertTrue(('subtable', ('foo', 3)) in rows)
        self.assertTrue(('subtable', ('bar', 'a')) in rows)
        self.assertTrue(('subtable', ('bar', 'b')) in rows)
        self.assertTrue(('testtable', ('foo',)) in rows)
        self.assertTrue(('testtable', ('bar',)) in rows)

    def test_convert_vdict_list_parent_key_with_id(self):
        # Test a VDICT that contains lists using a parent_key.
        resp = {'foo': (1, 2, 3), 'bar': ('a', 'b')}
        translator = {'translation-type': 'VDICT', 'table-name': 'testtable',
                      'id-col': 'id', 'key-col': 'key', 'val-col': 'value',
                      'translator': {'translation-type': 'LIST',
                                     'table-name': 'subtable',
                                     'parent-key': 'key',
                                     'val-col': 'val_col',
                                     'translator': self.val_trans}}
        rows, actual_k = datasource_driver.DataSourceDriver.convert_obj(
            resp, translator)

        k = self.compute_hash((('foo',), ('bar',)))

        self.assertEqual(7, len(rows))
        self.assertEqual(k, actual_k)

        self.assertTrue(('subtable', ('foo', 1)) in rows)
        self.assertTrue(('subtable', ('foo', 2)) in rows)
        self.assertTrue(('subtable', ('foo', 3)) in rows)
        self.assertTrue(('subtable', ('bar', 'a')) in rows)
        self.assertTrue(('subtable', ('bar', 'b')) in rows)
        self.assertTrue(('testtable', (k, 'foo')) in rows)
        self.assertTrue(('testtable', (k, 'bar')) in rows)

    def test_convert_bad_params(self):
        def verify_invalid_params(translator, err_msg):
            args = helper.datasource_openstack_args()
            driver = datasource_driver.DataSourceDriver('', '', None, None,
                                                        args=args)
            try:
                driver.register_translator(translator)
            except exception.InvalidParamException as e:
                self.assertTrue(err_msg in str(e))
            else:
                self.fail("Expected InvalidParamException but got none")

        # Test an invalid translation-type.
        verify_invalid_params(
            {'translation-typeXX': 'VDICT', 'table-name': 'testtable',
             'id-col': 'id_col', 'key-col': 'key', 'val-col': 'value',
             'translator': self.val_trans},
            'Param (translation-type) must be in translator')

        # Test invalid HDICT params
        verify_invalid_params(
            {'translation-type': 'HDICT', 'table-nameXX': 'testtable',
             'id-col': 'id_col', 'selector-type': 'DOT_SELECTOR',
             'field-translators': ({'fieldname': 'abc',
                                   'translator': self.val_trans},)},
            'Params (table-nameXX) are invalid')

        # Test invalid HDICT field translator params
        verify_invalid_params(
            {'translation-type': 'HDICT', 'table-name': 'testtable',
             'id-col': 'id_col', 'selector-type': 'DOT_SELECTOR',
             'field-translators':
                 ({'fieldname': 'abc',
                   'translator': {'translation-typeXX': 'VALUE'}},)},
            'Param (translation-type) must be in translator')

        # Test invalid HDICT field translator params
        verify_invalid_params(
            {'translation-type': 'HDICT', 'table-name': 'testtable',
             'id-col': 'id_col', 'selector-type': 'DOT_SELECTOR',
             'field-translators':
                 ({'fieldname': 'abc',
                   'translator': {'translation-type': 'VALUE',
                                  'XX': 123}},)},
            'Params (XX) are invalid')

        # Test invalid VDICT params
        verify_invalid_params(
            {'translation-type': 'VDICT', 'table-nameXX': 'testtable',
             'id-col': 'id_col', 'key-col': 'key', 'val-col': 'value',
             'translator': self.val_trans},
            'Params (table-nameXX) are invalid')

        # Test invalid VDICT sub translator params
        verify_invalid_params(
            {'translation-type': 'VDICT', 'table-name': 'testtable',
             'id-col': 'id_col', 'key-col': 'key', 'val-col': 'value',
             'translator': {'translation-typeXX': 'VALUE'}},
            'Param (translation-type) must be in translator')

        # Test invalid VDICT sub translator params
        verify_invalid_params(
            {'translation-type': 'VDICT', 'table-nameXX': 'testtable',
             'id-col': 'id_col', 'key-col': 'key', 'val-col': 'value',
             'translator': {'translation-type': 'VALUE', 'XX': 123}},
            'Params (table-nameXX) are invalid')

        # Test invalid LIST params
        verify_invalid_params(
            {'translation-type': 'LIST', 'table-nameXX': 'testtable',
             'id-col': 'id_col', 'val-col': 'value',
             'translator': self.val_trans},
            'Params (table-nameXX) are invalid')

        # Test invalid LIST sub translator params
        verify_invalid_params(
            {'translation-type': 'VDICT', 'table-name': 'testtable',
             'id-col': 'id_col', 'key-col': 'key', 'val-col': 'value',
             'translator': {'translation-typeXX': 'VALUE'}},
            'Param (translation-type) must be in translator')

        # Test invalid LIST sub translator params
        verify_invalid_params(
            {'translation-type': 'VDICT', 'table-name': 'testtable',
             'id-col': 'id_col', 'key-col': 'key', 'val-col': 'value',
             'translator': {'translation-type': 'VALUE', 'XX': 123}},
            'Params (XX) are invalid')

    def test_get_schema(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            translator = {
                'translation-type': 'HDICT',
                'table-name': 'testtable',
                'selector-type': 'DOT_SELECTOR',
                'field-translators': (
                    {'fieldname': 'testfield1',
                     'col': 'parent_col1',
                     'translator': {'translation-type': 'HDICT',
                                    'table-name': 'subtable1',
                                    'id-col': 'id1',
                                    'field-translators': (
                                        {'fieldname': 'a',
                                         'col': 'a1',
                                         'translator': self.val_trans},
                                        {'fieldname': 'b',
                                         'col': 'b1',
                                         'translator': self.val_trans})}},
                    {'fieldname': 'testfield2',
                     'translator': {'translation-type': 'HDICT',
                                    'table-name': 'subtable2',
                                    'id-col': 'id2',
                                    'field-translators': (
                                        {'fieldname': 'c',
                                         'col': 'c1',
                                         'translator': self.val_trans},
                                        {'fieldname': 'd',
                                         'col': 'd1',
                                         'translator': self.val_trans})}},
                    {'fieldname': 'ztestfield3', 'col': 'zparent_col3',
                     'translator': self.val_trans},
                    {'fieldname': 'testfield4', 'col': 'parent_col4',
                     'translator': {'translation-type': 'VALUE',
                                    'extract-fn': lambda x: x.id}},
                    {'fieldname': 'testfield5', 'col': 'parent_col5',
                     'translator': {'translation-type': 'VDICT',
                                    'table-name': 'subtable3', 'id-col': 'id3',
                                    'key-col': 'key3', 'val-col': 'value3',
                                    'translator': self.val_trans}},
                    {'fieldname': 'testfield6', 'col': 'parent_col6',
                     'translator': {'translation-type': 'LIST',
                                    'table-name': 'subtable4', 'id-col': 'id4',
                                    'val-col': 'value4',
                                    'translator': self.val_trans}},
                    {'fieldname': 'testfield7', 'col': 'parent_col7',
                     'translator': {'translation-type': 'VDICT',
                                    'table-name': 'subtable5',
                                    'key-col': 'key5', 'val-col': 'value5',
                                    'translator': self.val_trans}},
                    {'fieldname': 'testfield8', 'col': 'parent_col8',
                     'translator': {'translation-type': 'LIST',
                                    'table-name': 'subtable6',
                                    'val-col': 'value6',
                                    'translator': self.val_trans}})}

            TRANSLATORS = [translator]

            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        schema = TestDriver().get_schema()
        self.assertEqual(7, len(schema))

        self.assertEqual(({'name': 'id1', 'desc': None},
                          {'name': 'a1', 'desc': None},
                          {'name': 'b1', 'desc': None}), schema['subtable1'])
        self.assertEqual(({'name': 'id2', 'desc': None},
                          {'name': 'c1', 'desc': None},
                          {'name': 'd1', 'desc': None}), schema['subtable2'])
        self.assertEqual(('id3', 'key3', 'value3'), schema['subtable3'])
        self.assertEqual(
            ({'name': 'id4', 'desc': None},
             {'name': 'value4', 'desc': None}), schema['subtable4'])
        self.assertEqual(('key5', 'value5'), schema['subtable5'])
        self.assertEqual(({'name': 'value6',
                           'desc': None},), schema['subtable6'])
        self.assertEqual(
            ({'name': 'parent_col1', 'desc': None},
             {'name': 'testfield2', 'desc': None},
             {'name': 'zparent_col3', 'desc': None},
             {'name': 'parent_col4', 'desc': None},
             {'name': 'parent_col5', 'desc': None},
             {'name': 'parent_col6', 'desc': None},
             {'name': 'parent_col7', 'desc': None},
             {'name': 'parent_col8', 'desc': None}), schema['testtable'])

    def test_get_schema_with_table_reuse(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            translator = {'translation-type': 'LIST',
                          'table-name': 'testtable',
                          'id-col': 'id_col', 'val-col': 'value',
                          'translator': {'translation-type': 'LIST',
                                         'table-name': 'testtable',
                                         'id-col': 'id', 'val-col': 'val',
                                         'translator': self.val_trans}}

            TRANSLATORS = [translator]

            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        try:
            TestDriver().get_schema()
        except exception.DuplicateTableName as e:
            self.assertTrue('table (testtable) used twice' in str(e))
        else:
            self.fail("Expected InvalidParamException but got none")

    def test_get_schema_with_hdict_parent(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            subtranslator = {'translation-type': 'LIST',
                             'table-name': 'subtable',
                             'parent-key': 'id', 'val-col': 'val',
                             'translator': self.val_trans}

            translator = {'translation-type': 'HDICT',
                          'table-name': 'testtable',
                          'id-col': 'id_col',
                          'selector-type': 'DICT_SELECTOR',
                          'field-translators': ({'fieldname': 'unique_key',
                                                 'translator': self.val_trans},
                                                {'fieldname': 'sublist',
                                                 'translator': subtranslator})}

            TRANSLATORS = [translator]

            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        schema = TestDriver().get_schema()

        self.assertEqual(2, len(schema))
        self.assertEqual(
            ({'name': 'id_col', 'desc': None},
             {'name': 'unique_key', 'desc': None}), schema['testtable'])
        self.assertEqual(
            ({'name': 'parent_key', 'desc': None},
             {'name': 'val', 'desc': None}), schema['subtable'])

    def test_get_schema_with_hdict_id_function(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            translator = {
                'translation-type': 'HDICT',
                'table-name': 'testtable',
                'id-col': lambda obj: obj,
                'selector-type': 'DICT_SELECTOR',
                'field-translators': ({'fieldname': 'field1',
                                       'desc': 'test-field-1',
                                       'translator': self.val_trans},
                                      {'fieldname': 'field2',
                                       'desc': 'test-field-2',
                                       'translator': self.val_trans})}

            TRANSLATORS = [translator]

            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        schema = TestDriver().get_schema()

        self.assertEqual(1, len(schema))
        self.assertEqual(
            ({'name': 'id-col', 'desc': None},
             {'name': 'field1', 'desc': 'test-field-1'},
             {'name': 'field2', 'desc': 'test-field-2'}), schema['testtable'])

    def test_get_schema_with_vdict_parent(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            subtranslator = {'translation-type': 'LIST',
                             'table-name': 'subtable',
                             'parent-key': 'id_col', 'val-col': 'val',
                             'translator': self.val_trans}

            translator = {'translation-type': 'VDICT',
                          'table-name': 'testtable',
                          'id-col': 'id_col',
                          'key-col': 'key',
                          'val-col': 'val',
                          'translator': subtranslator}

            TRANSLATORS = [translator]

            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        schema = TestDriver().get_schema()

        self.assertEqual(2, len(schema))
        self.assertEqual(('id_col', 'key'), schema['testtable'])
        self.assertEqual(
            ({'name': 'parent_key', 'desc': None},
             {'name': 'val', 'desc': None}), schema['subtable'])

    def test_get_tablename(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            translator1 = {
                'translation-type': 'HDICT',
                'table-name': 'table-name1',
                'selector-type': 'DICT_SELECTOR',
                'field-translators':
                    ({'fieldname': 'col1', 'translator': self.val_trans},
                     {'fieldname': 'col2', 'translator': self.val_trans})
                }
            TRANSLATORS = [translator1]

            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        expected_ret = 'table-name1'
        ret = TestDriver().get_tablename('table-name1')
        self.assertEqual(expected_ret, ret)

    def test_get_tablenames(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            translator1 = {
                'translation-type': 'HDICT',
                'table-name': 'table-name1',
                'selector-type': 'DICT_SELECTOR',
                'field-translators':
                    ({'fieldname': 'col1', 'translator': self.val_trans},
                     {'fieldname': 'col2', 'translator': self.val_trans})
                }
            translator2 = {
                'translation-type': 'HDICT',
                'table-name': 'table-name2',
                'selector-type': 'DICT_SELECTOR',
                'field-translators':
                    ({'fieldname': 'col1', 'translator': self.val_trans},
                     {'fieldname': 'col2', 'translator': self.val_trans})
                }

            TRANSLATORS = [translator1, translator2]

            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        expected_ret = ['table-name1', 'table-name2']
        ret = TestDriver().get_tablenames()
        self.assertEqual(set(expected_ret), set(ret))

    def test_get_row_data(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        test_driver = TestDriver()
        test_driver.state = {'fake_table': [('d1', 'd2'), ('d3', 'd4')]}
        result = test_driver.get_row_data('fake_table')
        expected = [{'data': ('d1', 'd2')},
                    {'data': ('d3', 'd4')}]
        self.assertItemsEqual(expected, result)

    def test_nested_get_tables(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            translator2 = {
                'translation-type': 'HDICT',
                'table-name': 'table-name2',
                'selector-type': 'DICT_SELECTOR',
                'field-translators':
                    ({'fieldname': 'col1', 'translator': self.val_trans},
                     {'fieldname': 'col2', 'translator': self.val_trans})
                }

            translator1 = {
                'translation-type': 'HDICT',
                'table-name': 'table-name1',
                'selector-type': 'DICT_SELECTOR',
                'field-translators':
                    ({'fieldname': 'col1', 'translator': self.val_trans},
                     {'fieldname': 'col2', 'translator': translator2})
                }

            TRANSLATORS = [translator1]

            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        expected_ret = ['table-name1', 'table-name2']
        ret = TestDriver().get_tablenames()
        self.assertEqual(set(expected_ret), set(ret))

    def test_update_state_on_changed(self):
        mocked_self = mock.MagicMock()
        mocked_self.raw_state = dict()
        resource = 'fake_resource'

        @datasource_utils.update_state_on_changed(resource)
        def _translate_raw_data(_self, raw_data):
            return mock.sentinel.translated_data

        result = _translate_raw_data(mocked_self, mock.sentinel.raw_data)

        self.assertEqual(mock.sentinel.translated_data, result)
        self.assertEqual(mock.sentinel.raw_data,
                         mocked_self.raw_state[resource])
        mocked_self._update_state.assert_called_once_with(
            resource, mock.sentinel.translated_data)

        # raw data is not changed, don't translate anything.
        result = _translate_raw_data(mocked_self, mock.sentinel.raw_data)

        self.assertEqual([], result)
        self.assertEqual(mock.sentinel.raw_data,
                         mocked_self.raw_state[resource])
        mocked_self._update_state.assert_called_once_with(
            resource, mock.sentinel.translated_data)

    def test_update_state_on_changed_with_changed_raw_data(self):
        mocked_self = mock.MagicMock()
        mocked_self.raw_state = dict()
        resource = 'fake_resource'
        mocked_self.raw_state[resource] = mock.sentinel.last_data

        @datasource_utils.update_state_on_changed(resource)
        def _translate_raw_data(_self, raw_data):
            return mock.sentinel.translated_data

        result = _translate_raw_data(mocked_self, mock.sentinel.new_data)

        self.assertEqual(mock.sentinel.translated_data, result)
        self.assertEqual(mock.sentinel.new_data,
                         mocked_self.raw_state[resource])
        mocked_self._update_state.assert_called_once_with(
            resource, mock.sentinel.translated_data)

    def test_update_state_on_changed_with_empty_raw_data(self):
        mocked_self = mock.MagicMock()
        mocked_self.raw_state = dict()
        resource = 'fake_resource'
        mocked_self.raw_state[resource] = mock.sentinel.last_data

        @datasource_utils.update_state_on_changed(resource)
        def _translate_raw_data(_self, raw_data):
            return []

        result = _translate_raw_data(mocked_self, [])

        self.assertEqual([], result)
        self.assertEqual([], mocked_self.raw_state[resource])
        mocked_self._update_state.assert_called_once_with(resource, [])

    # The test case should be removed, once oslo-incubator bug/1499369 is
    # resolved.
    def test_update_state_on_changed_with_wrong_eq(self):
        class EqObject(object):
            def __eq__(self, other):
                return True

        mocked_self = mock.MagicMock()
        mocked_self.raw_state = dict()
        resource = 'fake_resource'
        cached_data = EqObject()
        mocked_self.raw_state[resource] = [cached_data]

        @datasource_utils.update_state_on_changed(resource)
        def _translate_raw_data(_self, raw_data):
            return []

        new_data = EqObject()
        _translate_raw_data(mocked_self, [new_data])
        mocked_self._update_state.assert_called_once_with(resource, [])
        self.assertIs(new_data, mocked_self.raw_state[resource][0])

    def test_update_state(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        test_driver = TestDriver()
        test_driver.state = {'fake_table': set(), 'foo_table': set(),
                             'unchanged_table': {mock.sentinel.data}}
        test_driver._table_deps = {'fake_table': ['fake_table', 'foo_table'],
                                   'unchanged_table': ['unchanged_table']}

        row_data = [('fake_table', mock.sentinel.data1),
                    ('fake_table', mock.sentinel.data2),
                    ('foo_table', mock.sentinel.data3)]
        expected_state = {'fake_table': {mock.sentinel.data1,
                                         mock.sentinel.data2},
                          'foo_table': {mock.sentinel.data3},
                          'unchanged_table': {mock.sentinel.data}}

        test_driver._update_state('fake_table', row_data)

        self.assertEqual(expected_state, test_driver.state)

    def test_update_state_with_undefined_table(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        test_driver = TestDriver()
        test_driver.state = {'fake_table': set(), 'foo_table': set()}
        test_driver._table_deps = {'fake_table': ['fake_table', 'foo_table']}

        row_data = [('fake_table', mock.sentinel.data1),
                    ('foo_table', mock.sentinel.data2),
                    ('undefined_table', mock.sentinel.data3)]
        expected_state = {'fake_table': {mock.sentinel.data1},
                          'foo_table': {mock.sentinel.data2}}

        test_driver._update_state('fake_table', row_data)

        self.assertEqual(expected_state, test_driver.state)

    def test_update_state_with_none_row_data(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        test_driver = TestDriver()
        test_driver.state = {'fake_table': {mock.sentinel.data1},
                             'foo_table': {mock.sentinel.data2}}
        test_driver._table_deps = {'fake_table': ['fake_table', 'foo_table']}

        expected_state = {'fake_table': set(), 'foo_table': set()}
        test_driver._update_state('fake_table', [])

        self.assertEqual(expected_state, test_driver.state)

    def test_update_state_with_part_none_row_data(self):
        class TestDriver(datasource_driver.DataSourceDriver):
            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)

        test_driver = TestDriver()
        test_driver.state = {'fake_table': set(),
                             'foo_table': {mock.sentinel.data3}}
        test_driver._table_deps = {'fake_table': ['fake_table', 'foo_table']}

        row_data = [('fake_table', mock.sentinel.data1),
                    ('fake_table', mock.sentinel.data2)]
        expected_state = {'fake_table': {mock.sentinel.data1,
                                         mock.sentinel.data2},
                          'foo_table': set()}

        test_driver._update_state('fake_table', row_data)

        self.assertEqual(expected_state, test_driver.state)

    def test_build_table_deps(self):
        level10_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level10',
            'parent-key': 'parent_key',
            'selector-type': 'DICT_SELECTOR',
            'in-list': True,
            'field-translators':
                ({'fieldname': 'level3_thing', 'translator': self.val_trans},)}

        level3_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level3',
            'parent-key': 'parent_key',
            'selector-type': 'DICT_SELECTOR',
            'in-list': True,
            'field-translators':
                ({'fieldname': 'level3_thing', 'translator': self.val_trans},)}

        level2_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level2',
            'parent-key': 'id',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'thing', 'translator': self.val_trans},
                 {'fieldname': 'level3',
                  'translator': level3_translator})}

        level1_translator = {
            'translation-type': 'HDICT',
            'table-name': 'level1',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'id', 'translator': self.val_trans},
                 {'fieldname': 'level2',
                  'translator': level2_translator})}

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        driver.register_translator(level1_translator)
        driver.register_translator(level10_translator)
        expected_table_deps = {'level1': ['level1', 'level2', 'level3'],
                               'level10': ['level10']}
        self.assertEqual(expected_table_deps, driver._table_deps)

    @mock.patch.object(eventlet, 'spawn')
    def test_init_consistence_with_exception(self, mock_spawn):
        class TestDriver(datasource_driver.DataSourceDriver):
            def __init__(self):
                super(TestDriver, self).__init__('', '', None, None, None)
                self.do_something()
                self._init_end_start_poll()

            def do_something(self):
                pass

        with mock.patch.object(TestDriver, "do_something",
                               side_effect=Exception()):
            test_driver = None
            try:
                test_driver = TestDriver()
                self.fail("Exception should be raised")
            except Exception:
                self.assertEqual(0, mock_spawn.call_count)
                self.assertIsNone(test_driver)

    def test_objects_extract_func(self):
        def translate_json_str_to_list(objs):
            result = []
            data_list = objs['result']
            for k, v in data_list.items():
                dict_obj = json.loads(v)
                for key, value in dict_obj.items():
                    obj = {
                        'key': key,
                        'value': value
                        }
                    result.append(obj)

            return result

        test_translator = {
            'translation-type': 'HDICT',
            'table-name': 'test',
            'selector-type': 'DICT_SELECTOR',
            'objects-extract-fn': translate_json_str_to_list,
            'field-translators':
                ({'fieldname': 'key', 'translator': self.val_trans},
                 {'fieldname': 'value', 'translator': self.val_trans})
            }

        objs = {
            "result": {
                "data1": """{"key1": "value1", "key2": "value2"}""",
                }
            }

        expected_ret = [('test', ('key1', 'value1')),
                        ('test', ('key2', 'value2'))]

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        driver.register_translator(test_translator)

        ret = driver.convert_objs(objs, test_translator)

        for row in ret:
            self.assertTrue(row in expected_ret)
            expected_ret.remove(row)
        self.assertEqual([], expected_ret)

    def test_recursive_objects_extract_func(self):
        def translate_json_str_to_list(objs):
            result = []
            data_str = objs['data']
            dict_list = json.loads(data_str)
            for key, value in dict_list.items():
                obj = {
                    'key': key,
                    'value': value
                    }
                result.append(obj)
            return result

        test_child_translator = {
            'translation-type': 'HDICT',
            'table-name': 'test-child',
            'parent-key': 'id',
            'parent-col-name': 'result',
            'selector-type': 'DICT_SELECTOR',
            'in-list': True,
            'objects-extract-fn': translate_json_str_to_list,
            'field-translators':
                ({'fieldname': 'key', 'translator': self.val_trans},
                 {'fieldname': 'value', 'translator': self.val_trans})
            }

        test_parent_translator = {
            'translation-type': 'HDICT',
            'table-name': 'test-parent',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'id', 'translator': self.val_trans},
                 {'fieldname': 'result', 'translator': test_child_translator})
            }

        expected_ret = [('test-parent', ('id-1', )),
                        ('test-child', ('id-1', 'key1', 'value1')),
                        ('test-child', ('id-1', 'key2', 'value2'))]

        objs = [{
            "id": "id-1",
            "result": {
                "data": """{"key1": "value1", "key2": "value2"}""",
                }
            }]

        driver = datasource_driver.DataSourceDriver('', '', None, None, None)
        driver.register_translator(test_parent_translator)

        ret = driver.convert_objs(objs, test_parent_translator)

        for row in ret:
            self.assertTrue(row in expected_ret)
            expected_ret.remove(row)
        self.assertEqual([], expected_ret)


class TestPollingDataSourceDriver(base.TestCase):
    class TestDriver(datasource_driver.PollingDataSourceDriver):
        def __init__(self):
            super(TestPollingDataSourceDriver.TestDriver, self).__init__(
                '', '', None, None, None)
            self._init_end_start_poll()

    def setUp(self):
        super(TestPollingDataSourceDriver, self).setUp()

    @mock.patch.object(eventlet, 'spawn')
    def test_init_consistence(self, mock_spawn):
        test_driver = TestPollingDataSourceDriver.TestDriver()
        mock_spawn.assert_called_once_with(test_driver.poll_loop,
                                           test_driver.poll_time)
        self.assertTrue(test_driver.initialized)
        self.assertIsNotNone(test_driver.worker_greenthread)

    @mock.patch.object(eventlet.greenthread, 'kill')
    @mock.patch.object(eventlet, 'spawn')
    def test_cleanup(self, mock_spawn, mock_kill):
        dummy_thread = dict()
        mock_spawn.return_value = dummy_thread

        test_driver = TestPollingDataSourceDriver.TestDriver()

        self.assertEqual(test_driver.worker_greenthread, dummy_thread)

        test_driver.cleanup()

        mock_kill.assert_called_once_with(dummy_thread)
        self.assertIsNone(test_driver.worker_greenthread)


class TestPushedDriver(base.TestCase):
    class TestDriver(datasource_driver.PushedDataSourceDriver):
        value_trans = {'translation-type': 'VALUE'}
        test_translator = {
            'translation-type': 'HDICT',
            'table-name': 'test_translator',
            'selector-type': 'DICT_SELECTOR',
            'field-translators':
                ({'fieldname': 'id', 'translator': value_trans},
                 {'fieldname': 'name', 'translator': value_trans},
                 {'fieldname': 'status', 'translator': value_trans})
            }

        TRANSLATORS = [test_translator]

        def __init__(self):
            super(TestPushedDriver.TestDriver, self).__init__('test-pushed',
                                                              '', None,
                                                              None, {})

    def setUp(self):
        super(TestPushedDriver, self).setUp()

    def test_init_push_driver(self):
        test_driver = TestPushedDriver.TestDriver()
        self.assertTrue(test_driver.initialized)

    @mock.patch.object(datasource_driver.DataSourceDriver, 'publish')
    def test_push_entire_data(self, mock_publish):
        test_driver = TestPushedDriver.TestDriver()
        obj = [
            {'id': 1, 'name': 'column1', 'status': 'up'},
            {'id': 2, 'name': 'column2', 'status': 'down'}
            ]
        test_driver.update_entire_data('test_translator', obj)
        expected_state = set([
            (1, 'column1', 'up'),
            (2, 'column2', 'down')])

        mock_publish.assert_called_with('test_translator',
                                        test_driver.state['test_translator'])
        self.assertEqual(expected_state, test_driver.state['test_translator'])

    def test_ensure_cleanup_pushdriver(self):
        test_driver = TestPushedDriver.TestDriver()
        test_driver.cleanup()


class TestExecutionDriver(base.TestCase):

    def setUp(self):
        super(TestExecutionDriver, self).setUp()
        self.exec_driver = datasource_driver.ExecutionDriver()

    def test_get_method_nested(self):
        class server(object):
            def nested_method(self):
                return True

        class NovaClient(object):
            def __init__(self):
                self.servers = server()

            def top_method(self):
                return True

        nova_client = NovaClient()
        method = self.exec_driver._get_method(nova_client,
                                              "servers.nested_method")
        self.assertTrue(method())

    def test_get_method_top(self):
        class NovaClient(object):
            def top_method(self):
                return True

        nova_client = NovaClient()
        method = self.exec_driver._get_method(nova_client, "top_method")
        self.assertTrue(method())

    def test_execute_api(self):
        class NovaClient(object):
            def action(self, arg1, arg2, arg3):
                return "arg1=%s arg2=%s arg3=%s" % (arg1, arg2, arg3)

        nova_client = NovaClient()
        arg = {"positional": ["value1", "value2"], "named": {"arg3": "value3"}}
        # it will raise exception if the method _execute_api failed to location
        # the api
        self.exec_driver._execute_api(nova_client, "action", arg)

    def test_get_actions_order_by_name(self):
        mock_methods = {'funcA': mock.MagicMock(),
                        'funcH': mock.MagicMock(),
                        'funcF': mock.MagicMock()}
        with mock.patch.dict(self.exec_driver.executable_methods,
                             mock_methods):
            action_list = self.exec_driver.get_actions().get('results')
            expected_list = copy.deepcopy(action_list)
            expected_list.sort(key=lambda item: item['name'])
            self.assertEqual(expected_list, action_list)

    def test_add_executable_client_methods(self):
        class FakeNovaClient(object):

            def _internal_action(self, arg1, arg2):
                """internal action with docs.

                :param arg1: internal test arg1
                :param arg2: internal test arg2
                """
                pass

            def action_no_doc(self, arg1, arg2):
                pass

            def action_doc(self, arg1, arg2):
                """action with docs.

                :param arg1: test arg1
                :param arg2: test arg2
                """
                pass

        expected_methods = {'action_doc': [[{'desc': 'arg1: test arg1',
                                             'name': 'arg1'},
                                            {'desc': 'arg2: test arg2',
                                             'name': 'arg2'}],
                                           'action with docs. '],
                            'action_no_doc': [[], '']}

        nova_client = FakeNovaClient()
        api_prefix = 'congress.tests.datasources.test_datasource_driver'
        self.exec_driver.add_executable_client_methods(nova_client, api_prefix)
        self.assertEqual(expected_methods, self.exec_driver.executable_methods)



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import collections

from oslo_log import log as logging
import six

from congress import exception

LOG = logging.getLogger(__name__)

DATABASE_POLICY_TYPE = 'database'
NONRECURSIVE_POLICY_TYPE = 'nonrecursive'
ACTION_POLICY_TYPE = 'action'
MATERIALIZED_POLICY_TYPE = 'materialized'
DELTA_POLICY_TYPE = 'delta'
DATASOURCE_POLICY_TYPE = 'datasource'


class Tracer(object):
    def __init__(self):
        self.expressions = []
        self.funcs = [LOG.debug]   # functions to call to trace

    def trace(self, table):
        self.expressions.append(table)

    def is_traced(self, table):
        return table in self.expressions or '*' in self.expressions

    def log(self, table, msg, *args, **kwargs):
        depth = kwargs.pop("depth", 0)
        if kwargs:
            raise TypeError("Unexpected keyword arguments: %s" % kwargs)
        if self.is_traced(table):
            for func in self.funcs:
                func(("| " * depth) + msg, *args)


class StringTracer(Tracer):
    def __init__(self):
        super(StringTracer, self).__init__()
        self.stream = six.moves.StringIO()
        self.funcs.append(self.string_output)

    def string_output(self, msg, *args):
        self.stream.write((msg % args) + "\n")

    def get_value(self):
        return self.stream.getvalue()


##############################################################################
# Logical Building Blocks
##############################################################################

class Proof(object):
    """A single proof.

    Differs semantically from Database's
    Proof in that this version represents a proof that spans rules,
    instead of just a proof for a single rule.
    """
    def __init__(self, root, children):
        self.root = root
        self.children = children

    def __str__(self):
        return self.str_tree(0)

    def str_tree(self, depth):
        s = " " * depth
        s += str(self.root)
        s += "\n"
        for child in self.children:
            s += child.str_tree(depth + 1)
        return s

    def leaves(self):
        if len(self.children) == 0:
            return [self.root]
        result = []
        for child in self.children:
            result.extend(child.leaves())
        return result


##############################################################################
# Events
##############################################################################

class EventQueue(object):
    def __init__(self):
        self.queue = collections.deque()

    def enqueue(self, event):
        self.queue.append(event)

    def dequeue(self):
        return self.queue.popleft()

    def __len__(self):
        return len(self.queue)

    def __str__(self):
        return "[" + ",".join([str(x) for x in self.queue]) + "]"


##############################################################################
# Abstract Theories
##############################################################################

class Theory(object):
    def __init__(self, name=None, abbr=None, schema=None, theories=None,
                 id=None, desc=None, owner=None, kind=None):
        self.schema = schema
        self.theories = theories
        self.kind = kind
        self.id = id
        self.desc = desc
        self.owner = owner

        self.tracer = Tracer()
        if name is None:
            self.name = repr(self)
        else:
            self.name = name
        if abbr is None:
            self.abbr = "th"
        else:
            self.abbr = abbr
        maxlength = 6
        if len(self.abbr) > maxlength:
            self.trace_prefix = self.abbr[0:maxlength]
        else:
            self.trace_prefix = self.abbr + " " * (maxlength - len(self.abbr))

    def set_id(self, id):
        self.id = id

    def initialize_tables(self, tablenames, facts):
        """initialize_tables

        Event handler for (re)initializing a collection of tables.  Clears
        tables befores assigning the new table content.

        @facts must be an iterable containing compile.Fact objects.
        """
        raise NotImplementedError

    def actual_events(self, events):
        """Returns subset of EVENTS that are not noops."""
        actual = []
        for event in events:
            if event.insert:
                if event.formula not in self:
                    actual.append(event)
            else:
                if event.formula in self:
                    actual.append(event)
        return actual

    def debug_mode(self):
        tr = Tracer()
        tr.trace('*')
        self.set_tracer(tr)

    def set_tracer(self, tracer):
        self.tracer = tracer

    def get_tracer(self):
        return self.tracer

    def log(self, table, msg, *args, **kwargs):
        msg = self.trace_prefix + ": " + msg
        self.tracer.log(table, msg, *args, **kwargs)

    def policy(self):
        """Return a list of the policy statements in this theory."""
        raise NotImplementedError()

    def content(self):
        """Return a list of the contents of this theory.

        Maybe rules and/or data. Note: do not change name to CONTENTS, as this
        is reserved for a dictionary of stuff used by TopDownTheory.
        """
        raise NotImplementedError()

    def tablenames(self, body_only=False, include_builtin=False,
                   include_modal=True):
        tablenames = set()
        for rule in self.policy():
            tablenames |= rule.tablenames(
                body_only=body_only, include_builtin=include_builtin,
                include_modal=include_modal)
        return tablenames

    def __str__(self):
        return "Theory %s" % self.name

    def content_string(self):
        return '\n'.join([str(p) for p in self.content()]) + '\n'

    def get_rule(self, ident):
        for p in self.policy():
            if hasattr(p, 'id') and str(p.id) == str(ident):
                return p
        raise exception.NotFound('rule_id %s  is not found.' % ident)

    def arity(self, tablename, modal=None):
        """Return the number of columns for the given tablename.

        TABLENAME is of the form <policy>:<table> or <table>.
        MODAL is the value of the modal operator.
        """
        return NotImplementedError



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.datalog import base as datalog_base
from congress.datalog import compile
from congress.policy_engines import agnostic
from congress.tests import base
from congress.tests import helper

LOG = logging.getLogger(__name__)

MAT_THEORY = 'classification test theory'
DB_THEORY = 'database test theory'


class TestRuntime(base.TestCase):
    """Tests for Runtime that are not specific to any theory."""

    def prep_runtime(self, code=None, msg=None, target=None):
        # compile source
        if msg is not None:
            LOG.debug(msg)
        if code is None:
            code = ""
        if target is None:
            target = MAT_THEORY
        run = agnostic.Runtime()
        run.create_policy(MAT_THEORY,
                          kind=datalog_base.MATERIALIZED_POLICY_TYPE)
        run.create_policy(DB_THEORY,
                          kind=datalog_base.DATABASE_POLICY_TYPE)
        # ensure inserts without target go to MAT_THEORY
        run.DEFAULT_THEORY = MAT_THEORY
        run.debug_mode()
        run.insert(code, target=target)
        return run

    def check_equal(self, actual_string, correct_string, msg):
        self.assertTrue(helper.datalog_equal(
            actual_string, correct_string, msg))

    def check_db(self, runtime, correct_string, msg):
        """Check that runtime.theory[DB_THEORY] is equal to CORRECT_STRING."""
        self.check_equal(runtime.theory[DB_THEORY].content_string(),
                         correct_string, msg)

    def check_class(self, runtime, correct_string, msg, tablenames=None):
        """Test MAT_THEORY.

        Check that runtime RUN.theory[MAT_THEORY] is
        equal to CORRECT_STRING.
        """
        actual = runtime.theory[MAT_THEORY].content(tablenames=tablenames)
        actual_string = " ".join(str(x) for x in actual)
        self.check_equal(actual_string, correct_string, msg)

    def showdb(self, run):
        LOG.debug("Resulting DB: %s",
                  run.theory[MAT_THEORY].database | run.theory[DB_THEORY])

    def test_database(self):
        """Test Database with insert/delete."""
        run = self.prep_runtime('')

        self.check_db(run, "", "Empty database on init")

        # set semantics, not bag semantics
        run.insert('r(1)', DB_THEORY)
        self.check_db(run, "r(1)", "Basic insert")
        run.insert('r(1)', DB_THEORY)
        self.check_db(run, "r(1)", "Duplicate insert")

        run.delete('r(1)', DB_THEORY)
        self.check_db(run, "", "Delete")
        run.delete('r(1)', DB_THEORY)
        self.check_db(run, "", "Delete from empty table")

    def test_error_checking(self):
        """Test error-checking on insertion of rules."""
        code = ("p(x) :- q(x)")
        run = self.prep_runtime(code)
        result = run.get_target(MAT_THEORY).policy()
        self.assertEqual(1, len(result))
        self.assertTrue(compile.parse1("p(x) :- q(x)") in result)

        # safety 1
        code = ("p(x) :- not q(x)")
        run = self.prep_runtime("", "** Safety 1 **")
        permitted, changes = run.insert(code, MAT_THEORY)
        self.assertFalse(permitted)

        # safety 2
        code = ("p(x) :- q(y)")
        run = self.prep_runtime("", "** Safety 2 **")
        permitted, changes = run.insert(code, MAT_THEORY)
        self.assertFalse(permitted)

        # TODO(thinrichs): weaken cross-policy recursion restriction
        #   so that we can include recursion within a single theory.
        # recursion into classification theory
        # code = ("p(x) :- p(x)")
        # run = self.prep_runtime("", "** Classification Recursion **")
        # permitted, changes = run.insert(code, MAT_THEORY)
        # self.assertTrue(permitted)

        # stratification into classification theory
        code = ("p(x) :- q(x), not p(x)")
        run = self.prep_runtime("", "** Classification Stratification **")
        permitted, changes = run.insert(code, MAT_THEORY)
        self.assertFalse(permitted)

    def test_basic(self):
        """Materialized Theory: test rule propagation."""
        code = ("q(x) :- p(x), r(x)")
        run = self.prep_runtime(
            code, "**** Materialized Theory: Basic propagation tests ****")
        run.insert('r(1)', MAT_THEORY)
        run.insert('p(1)', MAT_THEORY)
        self.check_class(run, "r(1) p(1) q(1)",
                         "Insert into base table with 1 propagation")

        run.delete('r(1)', MAT_THEORY)
        self.check_class(run, "p(1)",
                         "Delete from base table with 1 propagation")

        # multiple rules
        code = ("q(x) :- p(x), r(x)"
                "q(x) :- s(x)")
        run.insert('p(1)', MAT_THEORY)
        run.insert('r(1)', MAT_THEORY)
        self.check_class(run, "p(1) r(1) q(1)", "Insert: multiple rules")
        run.insert('s(1)', MAT_THEORY)
        self.check_class(run, "p(1) r(1) s(1) q(1)",
                         "Insert: duplicate conclusions")

    def test_short_body(self):
        code = ("q(x) :- p(x)")
        run = self.prep_runtime(
            code, "**** Materialized Theory: Body length 1 tests ****")

        run.insert('p(1)', MAT_THEORY)
        self.check_class(run, "p(1) q(1)", "Insert with body of size 1")
        self.showdb(run)
        run.delete('p(1)', MAT_THEORY)
        self.showdb(run)
        self.check_class(run, "", "Delete with body of size 1")

    def test_existentials(self):
        code = ("q(x) :- p(x), r(y)")
        run = self.prep_runtime(
            code,
            "**** Materialized Theory: Unary tables with existential ****")
        run.insert('p(1)', MAT_THEORY)
        run.insert('r(2)', MAT_THEORY)
        run.insert('r(3)', MAT_THEORY)
        self.showdb(run)
        self.check_class(run, "p(1) r(2) r(3) q(1)",
                         "Insert with unary table and existential")
        run.delete('r(2)', MAT_THEORY)
        self.showdb(run)
        self.check_class(run, "p(1) r(3) q(1)",
                         "Delete 1 with unary table and existential")
        run.delete('r(3)', MAT_THEORY)
        self.check_class(run, "p(1)",
                         "Delete all with unary table and existential")

    def test_nonmonadic(self):
        run = self.prep_runtime(
            "q(x) :- p(x,y)",
            "**** Materialized Theory: Multiple-arity table tests ****")

        run.insert('p(1,2)', MAT_THEORY)
        self.check_class(run, "p(1, 2) q(1)",
                         "Insert: existential variable in body of size 1")
        run.delete('p(1,2)', MAT_THEORY)
        self.check_class(run, "",
                         "Delete: existential variable in body of size 1")

        code = ("q(x) :- p(x,y), r(y,x)")
        run = self.prep_runtime(code)
        run.insert('p(1,2)', MAT_THEORY)
        run.insert('r(2,1)', MAT_THEORY)
        self.check_class(run, "p(1, 2) r(2, 1) q(1)",
                         "Insert: join in body of size 2")
        run.delete('p(1,2)', MAT_THEORY)
        self.check_class(run, "r(2, 1)",
                         "Delete: join in body of size 2")
        run.insert('p(1,2)', MAT_THEORY)
        run.insert('p(1,3)', MAT_THEORY)
        run.insert('r(3,1)', MAT_THEORY)
        self.check_class(
            run, "r(2, 1) r(3,1) p(1, 2) p(1, 3) q(1)",
            "Insert: multiple existential bindings for same head")

        run.delete('p(1,2)', MAT_THEORY)
        self.check_class(
            run, "r(2, 1) r(3,1) p(1, 3) q(1)",
            "Delete: multiple existential bindings for same head")

    def test_larger_join(self):
        code = ("q(x,v) :- p(x,y), r(y,z), s(z,w), t(w,v)")
        run = self.prep_runtime(code)
        run.insert('p(1, 10)', MAT_THEORY)
        run.insert('p(1, 20)', MAT_THEORY)
        run.insert('p(10, 100)', MAT_THEORY)
        run.insert('p(20, 200)', MAT_THEORY)
        run.insert('p(100, 1000)', MAT_THEORY)
        run.insert('p(200, 2000)', MAT_THEORY)
        run.insert('p(1000, 10000)', MAT_THEORY)
        run.insert('p(2000, 20000)', MAT_THEORY)
        run.insert('r(10, 100)', MAT_THEORY)
        run.insert('r(20, 200)', MAT_THEORY)
        run.insert('s(100, 1000)', MAT_THEORY)
        run.insert('s(200, 2000)', MAT_THEORY)
        run.insert('t(1000, 10000)', MAT_THEORY)
        run.insert('t(2000, 20000)', MAT_THEORY)
        code = ("p(1,10) p(1,20) p(10, 100) p(20, 200) p(100, 1000) "
                "p(200, 2000) p(1000, 10000) p(2000, 20000) "
                "r(10,100) r(20,200) s(100,1000) s(200,2000) "
                "t(1000, 10000) t(2000,20000) "
                "q(1,10000) q(1,20000)")
        self.check_class(run, code, "Insert: larger join")
        run.delete('t(1000, 10000)', MAT_THEORY)
        code = ("p(1,10) p(1,20) p(10, 100) p(20, 200) p(100, 1000)  "
                "p(200, 2000) p(1000, 10000) p(2000, 20000) r(10,100) "
                "r(20,200) s(100,1000) s(200,2000) t(2000,20000) "
                "q(1,20000)")
        self.check_class(run, code, "Delete: larger join")

    def test_self_join(self):
        code = ("q(x,y) :- p(x,z), p(z,y)")
        run = self.prep_runtime(code)
        run.insert('p(1,2)', MAT_THEORY)
        run.insert('p(1,3)', MAT_THEORY)
        run.insert('p(2, 4)', MAT_THEORY)
        run.insert('p(2, 5)', MAT_THEORY)
        self.check_class(
            run, 'p(1,2) p(1,3) p(2,4) p(2,5) q(1,4) q(1,5)',
            "Insert: self-join", tablenames=['p', 'q'])
        run.delete('p(2, 4)', MAT_THEORY)
        self.check_class(run, 'p(1,2) p(1,3) p(2,5) q(1,5)', '',
                         tablenames=['p', 'q'])

        code = ("q(x,z) :- p(x,y), p(y,z)")
        run = self.prep_runtime(code)
        run.insert('p(1, 1)', MAT_THEORY)
        self.check_class(
            run, 'p(1,1) q(1,1)', "Insert: self-join on same data",
            tablenames=['p', 'q'])

        code = ("q(x,w) :- p(x,y), p(y,z), p(z,w)")
        run = self.prep_runtime(code)
        run.insert('p(1, 1)', MAT_THEORY)
        run.insert('p(1, 2)', MAT_THEORY)
        run.insert('p(2, 2)', MAT_THEORY)
        run.insert('p(2, 3)', MAT_THEORY)
        run.insert('p(2, 4)', MAT_THEORY)
        run.insert('p(2, 5)', MAT_THEORY)
        run.insert('p(3, 3)', MAT_THEORY)
        run.insert('p(3, 4)', MAT_THEORY)
        run.insert('p(3, 5)', MAT_THEORY)
        run.insert('p(3, 6)', MAT_THEORY)
        run.insert('p(3, 7)', MAT_THEORY)
        code = ('p(1,1) p(1,2) p(2,2) p(2,3) p(2,4) p(2,5)'
                'p(3,3) p(3,4) p(3,5) p(3,6) p(3,7)'
                'q(1,1) q(1,2) q(2,2) q(2,3) q(2,4) q(2,5)'
                'q(3,3) q(3,4) q(3,5) q(3,6) q(3,7)'
                'q(1,3) q(1,4) q(1,5) q(1,6) q(1,7)'
                'q(2,6) q(2,7)')
        self.check_class(run, code, "Insert: larger self join",
                         tablenames=['p', 'q'])
        run.delete('p(1, 1)', MAT_THEORY)
        run.delete('p(2, 2)', MAT_THEORY)
        code = ('       p(1,2)        p(2,3) p(2,4) p(2,5)'
                'p(3,3) p(3,4) p(3,5) p(3,6) p(3,7)'
                '                     q(2,3) q(2,4) q(2,5)'
                'q(3,3) q(3,4) q(3,5) q(3,6) q(3,7)'
                'q(1,3) q(1,4) q(1,5) q(1,6) q(1,7)'
                'q(2,6) q(2,7)')
        self.check_class(run, code, "Delete: larger self join",
                         tablenames=['p', 'q'])

    def test_insert_order(self):
        """Test insert.

        Test that the order in which we change rules
        and data is irrelevant.
        """
        # was actual bug: insert data first, then
        #   insert rule with self-join
        code = ('q(1)'
                'p(x) :- q(x)')
        run = self.prep_runtime(code)
        self.check_class(run, 'p(1) q(1)', "Basic insert order")

        code = ('s(1)'
                'q(1,1)'
                'p(x,y) :- q(x,y), not r(x,y)'
                'r(x,y) :- s(x), s(y)')
        run = self.prep_runtime(code)
        self.check_class(run, 's(1) q(1,1) r(1,1)', "Self-join Insert order",
                         tablenames=['s', 'q', 'r'])

        code = ('q(1)'
                'p(x) :- q(x) '
                'r(x) :- p(x) ')
        run = self.prep_runtime(code)
        self.check_class(run, 'q(1) p(1) r(1)', "Multiple rule insert")
        run.delete('p(x) :- q(x)', MAT_THEORY)
        self.check_class(run, 'q(1)', "Deletion of rule")

    def test_value_types(self):
        """Test the different value types."""
        # string
        code = ("q(x) :- p(x), r(x)")
        run = self.prep_runtime(
            code, "**** Materialized Theory: String data type ****")

        run.insert('r("apple")', MAT_THEORY)
        self.check_class(run, 'r("apple")',
                         "String insert with no propagations")
        run.insert('r("apple")', MAT_THEORY)
        self.check_class(run, 'r("apple")',
                         "Duplicate string insert with no propagations")

        run.delete('r("apple")', MAT_THEORY)
        self.check_class(run, "", "Delete with no propagations")
        run.delete('r("apple")', MAT_THEORY)
        self.check_class(run, "", "Delete from empty table")

        run.insert('r("apple")', MAT_THEORY)
        run.insert('p("apple")', MAT_THEORY)
        self.check_class(run, 'r("apple") p("apple") q("apple")',
                         "String insert with 1 propagation")

        run.delete('r("apple")', MAT_THEORY)
        self.check_class(run, 'p("apple")', "String delete with 1 propagation")

        # float
        code = ("q(x) :- p(x), r(x)")
        run = self.prep_runtime(
            code, "**** Materialized Theory: Float data type ****")

        run.insert('r(1.2)', MAT_THEORY)
        self.check_class(run, 'r(1.2)', "String insert with no propagations")
        run.insert('r(1.2)', MAT_THEORY)
        self.check_class(run, 'r(1.2)',
                         "Duplicate string insert with no propagations")

        run.delete('r(1.2)', MAT_THEORY)
        self.check_class(run, "", "Delete with no propagations")
        run.delete('r(1.2)', MAT_THEORY)
        self.check_class(run, "", "Delete from empty table")

        run.insert('r(1.2)', MAT_THEORY)
        run.insert('p(1.2)', MAT_THEORY)
        self.check_class(run, 'r(1.2) p(1.2) q(1.2)',
                         "String self.insert with 1 propagation")

        run.delete('r(1.2)', MAT_THEORY)
        self.check_class(run, 'p(1.2)', "String delete with 1 propagation")

    def test_negation(self):
        """Test Materialized Theory negation."""
        # Unary, single join
        code = ("q(x) :- p(x), not r(x)")
        run = self.prep_runtime(code,
                                "**** Materialized Theory: Negation ****")

        run.insert('p(2)', MAT_THEORY)
        self.check_class(run, 'p(2) q(2)',
                         "Insert into positive literal with propagation")
        run.delete('p(2)', MAT_THEORY)
        self.check_class(run, '',
                         "Delete from positive literal with propagation")

        run.insert('r(2)', MAT_THEORY)
        self.check_class(run, 'r(2)',
                         "Insert into negative literal without propagation")
        run.delete('r(2)', MAT_THEORY)
        self.check_class(run, '',
                         "Delete from negative literal without propagation")

        run.insert('p(2)', MAT_THEORY)
        run.insert('r(2)', MAT_THEORY)
        self.check_class(run, 'p(2) r(2)',
                         "Insert into negative literal with propagation")

        run.delete('r(2)', MAT_THEORY)
        self.check_class(run, 'q(2) p(2)',
                         "Delete from negative literal with propagation")

        # Unary, multiple joins
        code = ("s(x) :- p(x), not r(x), q(y), not t(y)")
        run = self.prep_runtime(code, "Unary, multiple join")
        run.insert('p(1)', MAT_THEORY)
        run.insert('q(2)', MAT_THEORY)
        self.check_class(run, 'p(1) q(2) s(1)',
                         'Insert with two negative literals')

        run.insert('r(3)', MAT_THEORY)
        self.check_class(run, 'p(1) q(2) s(1) r(3)',
                         'Ineffectual insert with 2 negative literals')
        run.insert('r(1)', MAT_THEORY)
        self.check_class(
            run, 'p(1) q(2) r(3) r(1)',
            'Insert into existentially quantified negative literal '
            'with propagation. ')
        run.insert('t(2)', MAT_THEORY)
        self.check_class(
            run, 'p(1) q(2) r(3) r(1) t(2)',
            'Insert into negative literal producing extra blocker for proof.')
        run.delete('t(2)', MAT_THEORY)
        self.check_class(run, 'p(1) q(2) r(3) r(1)',
                         'Delete first blocker from proof')
        run.delete('r(1)', MAT_THEORY)
        self.check_class(run, 'p(1) q(2) r(3) s(1)',
                         'Delete second blocker from proof')

        # Non-unary
        code = ("p(x, v) :- q(x,z), r(z, w), not s(x, w), u(w,v)")
        run = self.prep_runtime(code, "Non-unary")
        run.insert('q(1, 2)', MAT_THEORY)
        run.insert('r(2, 3)', MAT_THEORY)
        run.insert('r(2, 4)', MAT_THEORY)
        run.insert('u(3, 5)', MAT_THEORY)
        run.insert('u(4, 6)', MAT_THEORY)
        self.check_class(
            run, 'q(1,2) r(2,3) r(2,4) u(3,5) u(4,6) p(1,5) p(1,6)',
            'Insert with non-unary negative literal')

        run.insert('s(1, 3)', MAT_THEORY)
        self.check_class(
            run, 'q(1,2) r(2,3) r(2,4) u(3,5) u(4,6) s(1,3) p(1,6)',
            'Insert into non-unary negative with propagation')

        run.insert('s(1, 4)', MAT_THEORY)
        self.check_class(
            run, 'q(1,2) r(2,3) r(2,4) u(3,5) u(4,6) s(1,3) s(1,4)',
            'Insert into non-unary with different propagation')

    def test_select(self):
        """Materialized Theory: test the SELECT event handler."""
        code = ("p(x, y) :- q(x), r(y)")
        run = self.prep_runtime(code, "**** Materialized Theory: Select ****")
        run.insert('q(1)', MAT_THEORY)
        # self.assertEqual('q(1)', run.select('q(x)'))
        run.insert('q(2)', MAT_THEORY)
        run.insert('r(1)', MAT_THEORY)
        run.insert('r(2)', MAT_THEORY)
        self.check_class(
            run, 'q(1) q(2) r(1) r(2) p(1,1) p(1,2) p(2,1) p(2,2)',
            'Prepare for select')
        self.check_equal(
            run.select('p(x,y)', MAT_THEORY),
            'p(1,1) p(1,2) p(2,1) p(2,2)',
            'Select: bound no args')
        self.check_equal(
            run.select('p(1,y)', MAT_THEORY),
            'p(1,1) p(1,2)',
            'Select: bound 1st arg')
        self.check_equal(
            run.select('p(x,2)', MAT_THEORY),
            'p(1,2) p(2,2)',
            'Select: bound 2nd arg')
        self.check_equal(
            run.select('p(1,2)', MAT_THEORY),
            'p(1,2)',
            'Select: bound 1st and 2nd arg')
        self.check_equal(
            run.select('query :- q(x), r(y)', MAT_THEORY),
            'query :- q(1), r(1)'
            'query :- q(1), r(2)'
            'query :- q(2), r(1)'
            'query :- q(2), r(2)',
            'Select: conjunctive query')

    def test_modify_rules(self):
        """Test rules modification.

        Test the functionality for adding and deleting
        rules *after* data has already been entered.
        """
        run = self.prep_runtime("", "Rule modification")
        run.insert("q(1) r(1) q(2) r(2)", MAT_THEORY)
        self.showdb(run)
        self.check_class(run, 'q(1) r(1) q(2) r(2)', "Installation")
        run.insert("p(x) :- q(x), r(x)", MAT_THEORY)
        self.check_class(
            run,
            'q(1) r(1) q(2) r(2) p(1) p(2)', 'Rule insert after data insert')
        run.delete("q(1)", MAT_THEORY)
        self.check_class(
            run,
            'r(1) q(2) r(2) p(2)', 'Delete after Rule insert with propagation')
        run.insert("q(1)", MAT_THEORY)
        run.delete("p(x) :- q(x), r(x)", MAT_THEORY)
        self.check_class(run, 'q(1) r(1) q(2) r(2)', "Delete rule")

    def test_recursion(self):
        """Materialized Theory: test recursion."""
        self.skipTest('Recursion not currently allowed')
        # TODO(thinrichs): weaken cross-policy recursion restriction
        #   so that we can include recursion within a single theory.
        run = self.prep_runtime('q(x,y) :- p(x,y)'
                                'q(x,y) :- p(x,z), q(z,y)')
        run.insert('p(1,2)', MAT_THEORY)
        run.insert('p(2,3)', MAT_THEORY)
        run.insert('p(3,4)', MAT_THEORY)
        run.insert('p(4,5)', MAT_THEORY)
        self.check_class(
            run, 'p(1,2) p(2,3) p(3,4) p(4,5)'
            'q(1,2) q(2,3) q(1,3) q(3,4) q(2,4) q(1,4) q(4,5) q(3,5) '
            'q(1,5) q(2,5)', 'Insert into recursive rules')
        run.delete('p(1,2)', MAT_THEORY)
        self.check_class(
            run, 'p(2,3) p(3,4) p(4,5)'
            'q(2,3) q(3,4) q(2,4) q(4,5) q(3,5) q(2,5)',
            'Delete from recursive rules')



# Copyright (c) 2015 Hewlett Packard. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import os
import retrying


@retrying.retry(stop_max_attempt_number=20, wait_fixed=1000)
def retry_check_function_return_value(f, expected_value, error_msg=None):
    """Check if function f returns expected value."""
    if not error_msg:
        error_msg = 'Expected value "%s" not found' % expected_value
    r = f()
    if r != expected_value:
        raise Exception(error_msg)


def retry_on_exception(f):
    """Decorator to retry on an exception."""
    def wrapper():
        try:
            return f()
        except Exception:
            return False
    return wrapper


def root_path():
    """Return path to root of source code."""
    x = os.path.realpath(__file__)
    x, y = os.path.split(x)  # drop "helper.py"
    x, y = os.path.split(x)  # drop "scenario"
    x, y = os.path.split(x)  # drop "tests"
    x, y = os.path.split(x)  # drop "congress_tempest_tests"
    return x



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json

import mock
from oslo_config import cfg
import webob

from congress.api import application
from congress.api import webservice
from congress.tests import base


class TestApiApplication(base.TestCase):

    def setUp(self):
        super(TestApiApplication, self).setUp()
        cfg.CONF.set_override('auth_strategy', 'noauth')

    def _check_data_model_exc_response(self, method, exc, response):
        self.assertEqual(response.status_code, exc.http_status_code,
                         'Correct %s HTTP error status' % method)
        body = json.loads(response.body.decode('utf-8'))
        self.assertEqual(body['error']['error_code'], exc.error_code,
                         'Correct %s error code in response body' % method)
        self.assertEqual(
            body['error']['message'], exc.description,
            'Correct %s description in response body' % method)
        self.assertEqual(body['error']['error_data'], exc.data,
                         'Correct %s error data in response body' % method)

    def test_data_model_exception(self):
        exc = webservice.DataModelException(1, "error1", [1, {'a': 'b'}], 409)
        model = webservice.SimpleDataModel("test")
        for method in [m for m in dir(model) if "_item" in m]:
            setattr(model, method, mock.Mock(side_effect=exc))

        resource_mgr = application.ResourceManager()
        app = application.ApiApplication(resource_mgr)

        collection_handler = webservice.CollectionHandler(r'/c', model)
        resource_mgr.register_handler(collection_handler)
        for method in ['GET', 'POST']:
            request = webob.Request.blank('/c', body='{}'.encode('utf-8'),
                                          method=method)
            response = app(request)
            self._check_data_model_exc_response(method, exc, response)

        element_handler = webservice.ElementHandler(r'/e', model)
        resource_mgr.register_handler(element_handler)
        for method in ['GET', 'PUT', 'PATCH', 'DELETE']:
            request = webob.Request.blank('/e', body='{}'.encode('utf-8'),
                                          method=method)
            response = app(request)
            self._check_data_model_exc_response(method, exc, response)

    def _check_base_exc_response(self, method, response, expected_status):
        self.assertEqual(response.status_code, expected_status,
                         'Correct %s HTTP error status' % method)
        body = json.loads(response.body.decode('utf-8'))
        self.assertEqual(body['error']['error_code'], expected_status,
                         'Correct %s error code in response body' % method)
        if expected_status == 500:
            description = "Internal server error"
        elif expected_status == 404:
            description = "The resouce could not be found."
        else:
            self.fail("Unsupported expected_status value.")

        self.assertEqual(
            body['error']['message'], description,
            'Correct %s description in response body' % method)

    def test__exception(self):
        model = webservice.SimpleDataModel("test")
        for method in [m for m in dir(model) if "_item" in m]:
            setattr(model, method, mock.Mock(side_effect=Exception()))

        resource_mgr = application.ResourceManager()
        app = application.ApiApplication(resource_mgr)

        collection_handler = webservice.CollectionHandler(r'/c', model)
        resource_mgr.register_handler(collection_handler)
        for method in ['GET', 'POST']:
            request = webob.Request.blank('/c', body='{}'.encode('utf-8'),
                                          method=method)
            response = app(request)
            self._check_base_exc_response(method, response, 500)

        element_handler = webservice.ElementHandler(r'/e', model)
        resource_mgr.register_handler(element_handler)
        for method in ['GET', 'PUT', 'PATCH', 'DELETE']:
            request = webob.Request.blank('/e', body='{}'.encode('utf-8'),
                                          method=method)
            response = app(request)
            self._check_base_exc_response(method, response, 500)

            # Tests that making a request to an invalid url returns 404.
            request = webob.Request.blank('/invalid',
                                          body='{}'.encode('utf-8'),
                                          method=method)
            response = app(request)
            self._check_base_exc_response(method, response, 404)



# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg

from congress import exception
from congress import harness
from congress.managers import datasource as datasource_manager
from congress.tests import base
from congress.tests import fake_datasource
from congress.tests import helper


class TestDataSourceManager(base.SqlTestCase):

    def setUp(self):
        super(TestDataSourceManager, self).setUp()
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource'])
        self.datasource_mgr = datasource_manager.DataSourceManager
        self.datasource_mgr.validate_configured_drivers()
        self.cage = harness.create(helper.root_path())

    def _get_datasource_request(self):
        return {'id': 'asdf',
                'name': 'aaron',
                'driver': '',
                'description': 'hello world!',
                'enabled': True,
                'type': None,
                'config': {}}

    def test_make_datasource_dict(self):
        req = self._get_datasource_request()
        result = self.datasource_mgr.make_datasource_dict(req)
        self.assertEqual(req, result)

    def test_validate_create_datasource_invalid_driver(self):
        req = self._get_datasource_request()
        self.assertRaises(exception.InvalidDriver,
                          self.datasource_mgr.validate_create_datasource,
                          req)

    def test_validate_create_datasource_invalid_config_invalid_options(self):
        req = self._get_datasource_request()
        req['driver'] = 'invalid_datasource'
        self.assertRaises(exception.InvalidDriver,
                          self.datasource_mgr.validate_create_datasource,
                          req)

    def test_validate_create_datasource_missing_config_options(self):
        req = self._get_datasource_request()
        req['driver'] = 'fake_datasource'
        # This is still missing some required options
        req['config'] = {'auth_url': '1234'}
        self.assertRaises(exception.MissingRequiredConfigOptions,
                          self.datasource_mgr.validate_create_datasource,
                          req)

    def test_add_datasource(self):
        req = self._get_datasource_request()
        req['driver'] = 'fake_datasource'
        req['config'] = {'auth_url': 'foo',
                         'username': 'armax',
                         'password': 'password',
                         'tenant_name': 'armax'}
        # let driver generate this for us.
        del req['id']
        result = self.datasource_mgr.add_datasource(req)
        for key, value in req.items():
            self.assertEqual(value, result[key])
        # TODO(thinrichs): test that ensure the DB, the policy engine,
        #   and the datasource manager are all in sync

    def test_get_datasouce(self):
        req = self._get_datasource_request()
        req['driver'] = 'fake_datasource'
        req['config'] = {'auth_url': 'foo',
                         'username': 'armax',
                         'password': 'password',
                         'tenant_name': 'armax'}
        # let driver generate this for us.
        del req['id']
        result = self.datasource_mgr.add_datasource(req)
        result = self.datasource_mgr.get_datasource(result['id'])
        for key, value in req.items():
            self.assertEqual(value, result[key])

    def test_get_datasources(self):
        req = self._get_datasource_request()
        req['driver'] = 'fake_datasource'
        req['name'] = 'datasource1'
        req['config'] = {'auth_url': 'foo',
                         'username': 'armax',
                         'password': 'password',
                         'tenant_name': 'armax'}
        # let driver generate this for us.
        del req['id']
        self.datasource_mgr.add_datasource(req)
        req['name'] = 'datasource2'
        self.datasource_mgr.add_datasource(req)
        result = self.datasource_mgr.get_datasources()

        req['name'] = 'datasource1'
        for key, value in req.items():
            self.assertEqual(value, result[0][key])

        req['name'] = 'datasource2'
        for key, value in req.items():
            self.assertEqual(value, result[1][key])

    def test_get_datasources_hide_secret(self):
        req = self._get_datasource_request()
        req['driver'] = 'fake_datasource'
        req['name'] = 'datasource1'
        req['config'] = {'auth_url': 'foo',
                         'username': 'armax',
                         'password': 'password',
                         'tenant_name': 'armax'}
        # let driver generate this for us.
        del req['id']
        self.datasource_mgr.add_datasource(req)
        req['name'] = 'datasource2'
        self.datasource_mgr.add_datasource(req)

        # Value will be set as <hidden>
        req['config']['password'] = "<hidden>"
        result = self.datasource_mgr.get_datasources(filter_secret=True)

        req['name'] = 'datasource1'
        for key, value in req.items():
            self.assertEqual(value, result[0][key])

        req['name'] = 'datasource2'
        for key, value in req.items():
            self.assertEqual(value, result[1][key])

    def test_create_datasource_duplicate_name(self):
        req = self._get_datasource_request()
        req['driver'] = 'fake_datasource'
        req['name'] = 'datasource1'
        req['config'] = {'auth_url': 'foo',
                         'username': 'armax',
                         'password': 'password',
                         'tenant_name': 'armax'}
        # let driver generate this for us.
        del req['id']
        self.datasource_mgr.add_datasource(req)
        self.assertRaises(exception.DatasourceNameInUse,
                          self.datasource_mgr.add_datasource, req)

    def test_delete_datasource(self):
        req = self._get_datasource_request()
        req['driver'] = 'fake_datasource'
        req['config'] = {'auth_url': 'foo',
                         'username': 'armax',
                         'password': 'password',
                         'tenant_name': 'armax'}
        # let driver generate this for us.
        del req['id']
        result = self.datasource_mgr.add_datasource(req)
        self.datasource_mgr.delete_datasource(result['id'])
        self.assertRaises(exception.DatasourceNotFound,
                          self.datasource_mgr.get_datasource,
                          result['id'])
        engine = self.cage.service_object('engine')
        self.assertRaises(exception.PolicyRuntimeException,
                          engine.assert_policy_exists, req['name'])
        # TODO(thinrichs): test that we've actually removed
        #   the row from the DB

    def test_delete_datasource_error(self):
        req = self._get_datasource_request()
        req['driver'] = 'fake_datasource'
        req['config'] = {'auth_url': 'foo',
                         'username': 'armax',
                         'password': 'password',
                         'tenant_name': 'armax'}
        # let driver generate this for us.
        del req['id']
        result = self.datasource_mgr.add_datasource(req)
        engine = self.cage.service_object('engine')
        engine.create_policy('alice')
        engine.insert('p(x) :- %s:q(x)' % req['name'], 'alice')
        self.assertRaises(exception.DanglingReference,
                          self.datasource_mgr.delete_datasource,
                          result['id'])

    def test_delete_invalid_datasource(self):
        self.assertRaises(exception.DatasourceNotFound,
                          self.datasource_mgr.delete_datasource,
                          "does_not_exist")

    def test_get_driver_schema(self):
        schema = self.datasource_mgr.get_driver_schema(
            'fake_datasource')
        self.assertEqual(
            schema,
            fake_datasource.FakeDataSource.get_schema())

    def test_get_datasouce_schema_driver_not_found(self):
        self.assertRaises(exception.DatasourceNotFound,
                          self.datasource_mgr.get_datasource_schema,
                          "does_not_exist")

    def test_duplicate_driver_name_raises(self):
        # Load the driver twice
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource',
             'congress.tests.fake_datasource.FakeDataSource'])
        self.datasource_mgr = datasource_manager.DataSourceManager
        self.assertRaises(exception.BadConfig,
                          self.datasource_mgr.validate_configured_drivers)

    def test_datasource_spawn_datasource_poll(self):
        req = self._get_datasource_request()
        req['driver'] = 'fake_datasource'
        req['config'] = {'auth_url': 'foo',
                         'username': 'armax',
                         'password': 'password',
                         'tenant_name': 'armax'}
        # let driver generate this for us.
        del req['id']
        result = self.datasource_mgr.add_datasource(req)
        self.datasource_mgr.request_refresh(result['id'])
        # TODO(thinrichs): test that the driver actually polls

    def test_datasource_spawn_datasource_poll_not_found(self):
        self.assertRaises(exception.DatasourceNotFound,
                          self.datasource_mgr.request_refresh,
                          "does_not_exist")



# Copyright 2014 VMware.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import logging

from django.core.urlresolvers import reverse
from django.template.defaultfilters import slugify
from django.utils.translation import ugettext_lazy as _
from horizon import exceptions
from horizon import messages
from horizon import tables
from openstack_dashboard.api import congress
from openstack_dashboard.dashboards.admin.datasources import (
    tables as datasources_tables)


logger = logging.getLogger(__name__)


class IndexView(tables.MultiTableView):
    """List service and policy defined data."""
    table_classes = (datasources_tables.DataSourcesTablesTable,
                     datasources_tables.PoliciesTablesTable,
                     datasources_tables.DataSourceStatusesTable,)
    template_name = 'admin/datasources/index.html'

    def get_datasources_tables_data(self):
        try:
            datasources = congress.datasources_list(self.request)
        except Exception as e:
            msg = _('Unable to get services list: %s') % e.message
            messages.error(self.request, msg)
            return []

        ds_temp = []
        for ds in datasources:
            ds_id = ds['id']
            try:
                ds_tables = congress.datasource_tables_list(self.request,
                                                            ds_id)
            except Exception as e:
                msg_args = {'ds_id': ds_id, 'error': e.message}
                msg = _('Unable to get tables list for service "%(ds_id)s": '
                        '%(error)s') % msg_args
                messages.error(self.request, msg)
                return []

            for table in ds_tables:
                table.set_value('datasource_id', ds_id)
                table.set_value('datasource_name', ds['name'])
                table.set_value('datasource_driver', ds['driver'])
                table.set_id_as_name_if_empty()
                # Object ids within a Horizon table must be unique. Otherwise,
                # Horizon will cache the column values for the object by id and
                # use the same column values for all rows with the same id.
                table.set_value('table_id', table['id'])
                table.set_value('id', '%s-%s' % (ds_id, table['table_id']))
                ds_temp.append(table)

        logger.debug("ds_temp %s" % ds_temp)
        return ds_temp

    def get_service_status_data(self):
        ds = []
        try:
            ds = congress.datasource_statuses_list(self.request)
            logger.debug("ds status : %s " % ds)
        except Exception as e:
            msg = _('Unable to get datasource status list: %s') % e.message
            messages.error(self.request, msg)
        return ds

    def get_policies_tables_data(self):
        try:
            policies = congress.policies_list(self.request)
        except Exception as e:
            msg = _('Unable to get policies list: %s') % e.message
            messages.error(self.request, msg)
            return []

        policies_tables = []
        for policy in policies:
            policy_name = policy['name']
            try:
                policy_tables = congress.policy_tables_list(self.request,
                                                            policy_name)
            except Exception as e:
                msg_args = {'policy_name': policy_name, 'error': e.message}
                msg = _('Unable to get tables list for policy '
                        '"%(policy_name)s": %(error)s') % msg_args
                messages.error(self.request, msg)
                return []

            for pt in policy_tables:
                pt.set_id_as_name_if_empty()
                pt.set_policy_details(policy)
                # Object ids within a Horizon table must be unique. Otherwise,
                # Horizon will cache the column values for the object by id and
                # use the same column values for all rows with the same id.
                pt.set_value('table_id', pt['id'])
                pt.set_value('id', '%s-%s' % (policy_name, pt['table_id']))
            policies_tables.extend(policy_tables)

        return policies_tables


class DetailView(tables.DataTableView):
    """List details about and rows from a data source (service or policy)."""
    table_class = datasources_tables.DataSourceRowsTable
    template_name = 'admin/datasources/detail.html'

    def get_data(self):
        datasource_id = self.kwargs['datasource_id']
        table_name = self.kwargs.get('policy_table_name')
        is_service = False

        try:
            if table_name:
                # Policy data table.
                rows = congress.policy_rows_list(self.request, datasource_id,
                                                 table_name)
                if congress.TABLE_SEPARATOR in table_name:
                    table_name_parts = table_name.split(
                        congress.TABLE_SEPARATOR)
                    maybe_datasource_name = table_name_parts[0]
                    datasources = congress.datasources_list(self.request)
                    for datasource in datasources:
                        if datasource['name'] == maybe_datasource_name:
                            # Serivce-derived policy data table.
                            is_service = True
                            datasource_id = datasource['id']
                            table_name = table_name_parts[1]
                            break
            else:
                # Service data table.
                is_service = True
                datasource = congress.datasource_get_by_name(
                    self.request, datasource_id)
                table_name = self.kwargs['service_table_name']
                rows = congress.datasource_rows_list(
                    self.request, datasource_id, table_name)
        except Exception as e:
            msg_args = {
                'table_name': table_name,
                'ds_id': datasource_id,
                'error': e.message
            }
            msg = _('Unable to get rows in table "%(table_name)s", data '
                    'source "%(ds_id)s": %(error)s') % msg_args
            messages.error(self.request, msg)
            redirect = reverse('horizon:admin:datasources:index')
            raise exceptions.Http302(redirect)

        # Normally, in Horizon, the columns for a table are defined as
        # attributes of the Table class. When the class is instantiated,
        # the columns are processed during the metaclass initialization. To
        # add columns dynamically, re-create the class from the metaclass
        # with the added columns, re-create the Table from the new class,
        # then reassign the Table stored in this View.
        column_names = []
        table_class_attrs = copy.deepcopy(dict(self.table_class.__dict__))
        # Get schema from the server.
        try:
            if is_service:
                schema = congress.datasource_table_schema_get(
                    self.request, datasource_id, table_name)
            else:
                schema = congress.policy_table_schema_get(
                    self.request, datasource_id, table_name)
        except Exception as e:
            msg_args = {
                'table_name': table_name,
                'ds_id': datasource_id,
                'error': e.message
            }
            msg = _('Unable to get schema for table "%(table_name)s", '
                    'data source "%(ds_id)s": %(error)s') % msg_args
            messages.error(self.request, msg)
            redirect = reverse('horizon:admin:datasources:index')
            raise exceptions.Http302(redirect)

        columns = schema['columns']
        row_len = 0
        if len(rows):
            row_len = len(rows[0].get('data', []))

        if not row_len or row_len == len(columns):
            for col in columns:
                col_name = col['name']
                # Attribute name for column in the class must be a valid
                # identifier. Slugify it.
                col_slug = slugify(col_name)
                column_names.append(col_slug)
                table_class_attrs[col_slug] = tables.Column(
                    col_slug, verbose_name=col_name)
        else:
            # There could be another table with the same name and different
            # arity. Divide the rows into unnamed columns. Number them for
            # internal reference.
            for i in xrange(0, row_len):
                col_name = str(i)
                column_names.append(col_name)
                table_class_attrs[col_name] = tables.Column(
                    col_name, verbose_name='')

        # Class and object re-creation, using a new class name, the same base
        # classes, and the new class attributes, which now includes columns.
        columnized_table_class_name = '%s%sRows' % (
            slugify(datasource_id).title(), slugify(table_name).title())
        columnized_table_class = tables.base.DataTableMetaclass(
            str(columnized_table_class_name), self.table_class.__bases__,
            table_class_attrs)

        self.table_class = columnized_table_class
        columnized_table = columnized_table_class(self.request, **self.kwargs)
        self._tables[columnized_table_class._meta.name] = columnized_table

        # Map columns names to row values.
        num_cols = len(column_names)
        for row in rows:
            try:
                row_data = row['data']
                row.delete_by_key('data')
                for i in xrange(0, num_cols):
                    row.set_value(column_names[i], row_data[i])
            except Exception as e:
                msg_args = {
                    'table_name': table_name,
                    'ds_id': datasource_id,
                    'error': e.message
                }
                msg = _('Unable to get data for table "%(table_name)s", data '
                        'source "%(ds_id)s": %(error)s') % msg_args
                messages.error(self.request, msg)
                redirect = reverse('horizon:admin:datasources:index')
                raise exceptions.Http302(redirect)

        return rows

    def get_context_data(self, **kwargs):
        context = super(DetailView, self).get_context_data(**kwargs)
        if 'policy_table_name' in kwargs:
            table_name = kwargs.get('policy_table_name')
            context['datasource_type'] = _('Policy')
            datasource_name = kwargs['datasource_id']
        else:
            table_name = kwargs['service_table_name']
            context['datasource_type'] = _('Service')
            try:
                datasource_id = kwargs['datasource_id']
                datasource = congress.datasource_get(self.request,
                                                     datasource_id)
                datasource_name = datasource['name']
            except Exception as e:
                datasource_name = datasource_id
                logger.info('Failed to get data source "%s": %s' %
                            (datasource_id, e.message))
        context['datasource_name'] = datasource_name
        context['table_name'] = table_name
        return context



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg

from congress.api.system import driver_model
from congress.api import webservice
from congress import harness
from congress.managers import datasource as datasource_manager
from congress.tests import base
from congress.tests import helper


class TestDriverModel(base.SqlTestCase):
    def setUp(self):
        super(TestDriverModel, self).setUp()
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource'])

        self.cage = harness.create(helper.root_path())
        self.datasource_mgr = datasource_manager.DataSourceManager
        self.datasource_mgr.validate_configured_drivers()
        req = {'driver': 'fake_datasource',
               'name': 'fake_datasource'}
        req['config'] = {'auth_url': 'foo',
                         'username': 'foo',
                         'password': 'password',
                         'tenant_name': 'foo'}
        self.datasource = self.datasource_mgr.add_datasource(req)
        self.api_system = self.cage.service_object('api-system')
        self.driver_model = (
            driver_model.DatasourceDriverModel(
                "driver-model", {}, datasource_mgr=self.datasource_mgr)
        )

    def tearDown(self):
        super(TestDriverModel, self).tearDown()

    def test_drivers_list(self):
        context = {}
        expected_ret = {"results": [
            {
                "description": "This is a fake driver used for testing",
                "id": "fake_datasource"
            }
        ]}

        ret = self.driver_model.get_items({}, context)
        self.assertEqual(expected_ret, ret)

    def test_driver_details(self):
        context = {
            "driver_id": "fake_datasource"
        }
        expected_ret = {
            "config": {
                "auth_url": "required",
                "endpoint": "(optional)",
                "password": "required",
                "poll_time": "(optional)",
                "region": "(optional)",
                "project_name": "(optional)",
                "tenant_name": "required",
                "username": "required"
            },
            "description": "This is a fake driver used for testing",
            "id": "fake_datasource",
            "module": "congress.tests.fake_datasource.FakeDataSource",
            "secret": ["password"],
            "tables": [{'columns': [
                {'description': None, 'name': 'id'},
                {'description': None, 'name': 'name'}],
                'table_id': 'fake_table'}
            ]
        }

        ret = self.driver_model.get_item('fake_datasource', {}, context)
        self.assertEqual(expected_ret, ret)

    def test_invalid_driver_details(self):
        context = {
            "driver_id": "invalid-id"
        }
        self.assertRaises(webservice.DataModelException,
                          self.driver_model.get_item,
                          'invalid-id', {}, context)



# Copyright (c) 2016 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import uuid

from oslo_config import cfg
cfg.CONF.distributed_architecture = True

from congress.api import policy_model
from congress.api import rule_model
from congress.api import status_model
from congress.api import webservice
from congress.tests import base
from congress.tests2.api import base as api_base


class TestStatusModel(base.SqlTestCase):
    def setUp(self):
        super(TestStatusModel, self).setUp()
        self.policy_model = policy_model.PolicyModel('api-policy',
                                                     policy_engine='engine')
        self.rule_model = rule_model.RuleModel('api-rule',
                                               policy_engine='engine')
        self.status_model = status_model.StatusModel('api-status',
                                                     policy_engine='engine')
        result = api_base.setup_config([self.policy_model, self.rule_model,
                                        self.status_model])
        self.node = result['node']
        self.datasource = result['data']

    def test_get_datasource_status(self):
        context = {'ds_id': self.datasource.service_id}
        status = self.status_model.get_item(None, {}, context=context)
        expected_status_keys = ['last_updated', 'subscriptions',
                                'last_error', 'subscribers',
                                'initialized', 'number_of_updates']
        self.assertEqual(set(expected_status_keys), set(status.keys()))

    def test_get_invalid_datasource_status(self):
        context = {'ds_id': 'invalid_id'}
        self.assertRaises(webservice.DataModelException,
                          self.status_model.get_item, None, {},
                          context=context)

    def test_policy_id_status(self):
        result = self.policy_model.add_item({'name': 'test_policy'}, {})

        context = {'policy_id': result[0]}
        status = self.status_model.get_item(None, {}, context=context)
        expected_status = {'name': 'test_policy',
                           'id': result[0]}
        self.assertEqual(expected_status, status)

        # test with policy_name
        context = {'policy_id': result[1]['name']}
        status = self.status_model.get_item(None, {}, context=context)
        self.assertEqual(expected_status, status)

    def test_invalid_policy_id_status(self):
        invalid_id = uuid.uuid4()
        context = {'policy_id': invalid_id}
        self.assertRaises(webservice.DataModelException,
                          self.status_model.get_item, None, {},
                          context=context)

    def test_rule_status_policy_id(self):
        result = self.policy_model.add_item({'name': 'test_policy'}, {})
        policy_id = result[0]
        policy_name = result[1]['name']

        result = self.rule_model.add_item({'name': 'test_rule',
                                           'rule': 'p(x) :- q(x)'}, {},
                                          context={'policy_id': 'test_policy'})

        context = {'policy_id': policy_id, 'rule_id': result[0]}
        status = self.status_model.get_item(None, {}, context=context)
        expected_status = {'name': 'test_rule',
                           'id': result[0],
                           'comment': '',
                           'original_str': 'p(x) :- q(x)'}
        self.assertEqual(expected_status, status)

        # test with policy_name
        context = {'policy_id': policy_name, 'rule_id': result[0]}
        status = self.status_model.get_item(None, {}, context=context)
        expected_status = {'name': 'test_rule',
                           'id': result[0],
                           'comment': '',
                           'original_str': 'p(x) :- q(x)'}
        self.assertEqual(expected_status, status)

    def test_rule_status_invalid_rule_policy_id(self):
        result = self.policy_model.add_item({'name': 'test_policy'}, {})
        policy_id = result[0]
        invalid_rule = uuid.uuid4()

        context = {'policy_id': policy_id, 'rule_id': invalid_rule}
        self.assertRaises(webservice.DataModelException,
                          self.status_model.get_item, None, {},
                          context=context)

    def test_rule_status_invalid_policy_id(self):
        invalid_policy = uuid.uuid4()
        invalid_rule = uuid.uuid4()

        context = {'policy_id': invalid_policy, 'rule_id': invalid_rule}
        self.assertRaises(webservice.DataModelException,
                          self.status_model.get_item, None, {},
                          context=context)



# Copyright (c) 2014 Montavista Software, LLC.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging
import swiftclient.service

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return SwiftDriver(name, keys, inbox, datapath, args)


class SwiftDriver(datasource_driver.PollingDataSourceDriver,
                  datasource_driver.ExecutionDriver):

    CONTAINERS = "containers"
    OBJECTS = "objects"

    value_trans = {'translation-type': 'VALUE'}

    containers_translator = {
        'translation-type': 'HDICT',
        'table-name': CONTAINERS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'count', 'translator': value_trans},
             {'fieldname': 'bytes', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans})}

    objects_translator = {
        'translation-type': 'HDICT',
        'table-name': OBJECTS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'bytes', 'translator': value_trans},
             {'fieldname': 'last_modified', 'translator': value_trans},
             {'fieldname': 'hash', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'content_type', 'translator': value_trans},
             {'fieldname': 'container_name', 'translator': value_trans})}

    TRANSLATORS = [containers_translator, objects_translator]

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        if args is None:
            args = self.empty_credentials()
        super(SwiftDriver, self).__init__(name, keys, inbox, datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        options = self.get_swift_credentials_v1(args)
        self.swift_service = swiftclient.service.SwiftService(options)
        self.add_executable_client_methods(self.swift_service,
                                           'swiftclient.service')
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        # TODO(zhenzanz): This is verified with keystoneauth for swift.
        # Do we need to support other Swift auth systems?
        # http://docs.openstack.org/developer/swift/overview_auth.html
        result = {}
        result['id'] = 'swift'
        result['description'] = ('Datasource driver that interfaces with '
                                 'swift.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result

    def get_swift_credentials_v1(self, creds):
        # Check swiftclient/service.py _default_global_options for more
        # auth options. But these 4 options seem to be enough.
        options = {}
        options['os_username'] = creds['username']
        options['os_password'] = creds['password']
        options['os_tenant_name'] = creds['tenant_name']
        options['os_auth_url'] = creds['auth_url']
        return options

    def update_from_datasource(self):
        '''Read and populate.

        Read data from swift and populate the policy engine
        tables with current state as specified by translators
        '''
        containers, objects = self._get_containers_and_objects()

        LOG.debug("Containers Lists--->: %s" % containers)
        LOG.debug("Object Lists--->: %s " % objects)
        self._translate_containers(containers)
        self._translate_objects(objects)
        LOG.debug("CONTAINERS: %s" % str(self.state[self.CONTAINERS]))
        LOG.debug("OBJECTS: %s" % str(self.state[self.OBJECTS]))

    def _get_containers_and_objects(self):
        container_list = self.swift_service.list()
        cont_list = []
        objects = []
        containers = []
        LOG.debug("Swift obtaining containers List")
        for stats in container_list:
            containers = stats['listing']
            for item in containers:
                cont_list.append(item['name'])
        LOG.debug("Swift obtaining objects List")
        for container in cont_list:
            object_list = self.swift_service.list(container)
            for items in object_list:
                item_list = items['listing']
                for obj in item_list:
                    obj['container_name'] = container
                for obj in item_list:
                    objects.append(obj)
        return containers, objects

    @ds_utils.update_state_on_changed(CONTAINERS)
    def _translate_containers(self, obj):
        """Translate the containers represented by OBJ into tables."""
        row_data = SwiftDriver.convert_objs(obj,
                                            self.containers_translator)
        return row_data

    @ds_utils.update_state_on_changed(OBJECTS)
    def _translate_objects(self, obj):
        """Translate the objects represented by OBJ into tables."""
        row_data = SwiftDriver.convert_objs(obj,
                                            self.objects_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.swift_service, action, action_args)



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.datasources import datasource_driver


LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """Create dataservice instance.

    This method is called by d6cage to create a dataservice
    instance.  There are a couple of parameters we found useful
    to add to that call, so we included them here instead of
    modifying d6cage (and all the d6cage.createservice calls).
    """
    return TestDriver(name, keys, inbox, datapath, args)


class TestDriver(datasource_driver.PollingDataSourceDriver):
    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        if args is None:
            args = self._empty_openstack_credentials()
        super(TestDriver, self).__init__(name, keys, inbox, datapath, args)
        self.msg = None
        self.state = {}
        self._init_end_start_poll()

    def receive_msg(self, msg):
        LOG.info("TestDriver: received msg %s", msg)
        self.msg = msg

    def get_msg_data(self):
        msgstr = ""
        if self.msg is None:
            return msgstr
        # only support list and set now
        if isinstance(self.msg.body.data, (list, set)):
            for di in self.msg.body.data:
                msgstr += str(di)
        else:
            msgstr = str(self.msg.body.data)
        LOG.info("TestDriver: current received msg: %s", msgstr)
        return msgstr

    def update_from_datasource(self):
        pass

    def prepush_processor(self, data, dataindex, type=None):
        # don't change data before transfer
        return data



# Copyright (c) 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return FakeDataSource(name, keys, inbox, datapath, args)


class FakeDataSource(datasource_driver.PollingDataSourceDriver,
                     datasource_driver.PushedDataSourceDriver,
                     datasource_driver.ExecutionDriver):

    value_trans = {'translation-type': 'VALUE'}
    fake_translator = {
        'translation-type': 'HDICT',
        'table-name': 'fake_table',
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans})}

    TRANSLATORS = [fake_translator]

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        super(FakeDataSource, self).__init__(name, keys, inbox,
                                             datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.add_executable_method('fake_act',
                                   [{'name': 'server_id',
                                    'description': 'server to act'}],
                                   'fake action')
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'fake_datasource'
        result['description'] = 'This is a fake driver used for testing'
        result['config'] = datasource_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result

    def update_from_datasource(self):
        LOG.info("fake:: update_from_datasource")
